# ----------------------------------------------------------------------------------------------------------------------
#
# Copyright (C) 2019 Anton Synytsia
#
# ----------------------------------------------------------------------------------------------------------------------

unless file_loaded?(__FILE__)

file_loaded(__FILE__)

translate = AMS::ClothWorks.translate
cw_gdict = AMS::ClothWorks::CW_GDICT

# Choose icons path
simg_path = 'images/'
limg_path = 'images/'

# Context variables
licence_warning_displayed = false

# Create toolbar
toolbar = ::UI::Toolbar.new(translate['ClothWorks'])

# Create plugin menu
plugin_menu = ::UI.menu('Plugins').add_submenu(translate['ClothWorks'])

cmd = ::UI::Command.new('AMS CW Toggle UI') {
  if AMS::ClothWorks::UI.open?
    AMS::ClothWorks::UI.close
  else
    AMS::ClothWorks::UI.open
  end
}
cmd.set_validation_proc {
  next MF_GRAYED unless ::Sketchup.active_model
  AMS::ClothWorks::UI.open? ? MF_CHECKED : MF_UNCHECKED
}
cmd.menu_text = cmd.tooltip = translate['Toggle UI']
cmd.status_bar_text = translate['Open/close ClothWorks user interface.']
cmd.small_icon = simg_path + 'ui.png'
cmd.large_icon = limg_path + 'ui.png'
toolbar.add_item(cmd)
plugin_menu.add_item(cmd)

cmd = ::UI::Command.new('AMS CW Play') {
  if AMS::ClothWorks.simulation_active?
    AMS::ClothWorks.toggle_play_simulation
  else
    model = ::Sketchup.active_model

    num_sel_cloths = 0
    model.selection.each { |e|
      if (e.is_a?(::Sketchup::Group) || e.is_a?(::Sketchup::ComponentInstance)) && e.parent.is_a?(::Sketchup::Model) && e.valid? && e.visible? && e.layer.visible?
        type = e.get_attribute(cw_gdict, 'Type', nil)
        num_sel_cloths += 1 if type == 1
      end
    }

    num_cloths = 0
    model.entities.each { |e|
      if (e.is_a?(::Sketchup::Group) || e.is_a?(::Sketchup::ComponentInstance)) && e.valid? && e.visible? && e.layer.visible?
        type = e.get_attribute(cw_gdict, 'Type', nil)
        num_cloths += 1 if type == 1
      end
    }

    bss = false
    if num_sel_cloths > 0 && num_sel_cloths != num_cloths
      msg = translate["Would you like to simulate just the selected cloth?"]
      msg << "\n\n"
      msg << translate["Press YES to simulate selection; press NO to simulate all."]
      res = ::UI.messagebox(msg, MB_YESNO)
      if res == IDYES
        bss = true
      end
    end

    AMS::ClothWorks.start_simulation(model, bss)
  end
}
cmd.set_validation_proc {
  next MF_GRAYED unless ::Sketchup.active_model
  if AMS::ClothWorks.simulation_active?
    AMS::ClothWorks.simulation_instance.paused? ? MF_UNCHECKED : MF_CHECKED
  else
    MF_ENABLED
  end
}
cmd.menu_text = cmd.tooltip = translate['Toggle Play']
cmd.status_bar_text = translate['Play/pause ClothWorks simulation.']
cmd.small_icon = simg_path + 'play.png'
cmd.large_icon = limg_path + 'play.png'
toolbar.add_item(cmd)
plugin_menu.add_item(cmd)

cmd = ::UI::Command.new('AMS CW Stop') {
  AMS::ClothWorks.stop_simulation
}
cmd.set_validation_proc {
  next MF_GRAYED unless ::Sketchup.active_model
  AMS::ClothWorks.simulation_active? ? MF_ENABLED : MF_GRAYED
}
cmd.menu_text = cmd.tooltip = translate['Stop']
cmd.status_bar_text = translate['Stop the ClothWorks simulation. This option does not reset the cloth layout.']
cmd.small_icon = simg_path + 'stop.png'
cmd.large_icon = limg_path + 'stop.png'
toolbar.add_item(cmd)
plugin_menu.add_item(cmd)

cmd = ::UI::Command.new('AMS CW Reset') {
  AMS::ClothWorks.reset_simulation
}
cmd.set_validation_proc {
  next MF_GRAYED unless ::Sketchup.active_model
  AMS::ClothWorks.simulation_active? ? MF_ENABLED : MF_GRAYED
}
cmd.menu_text = cmd.tooltip = translate['Reset']
cmd.status_bar_text = translate['Reset the ClothWorks simulation. This option resets the cloth layout.']
cmd.small_icon = simg_path + 'reset.png'
cmd.large_icon = limg_path + 'reset.png'
toolbar.add_item(cmd)
plugin_menu.add_item(cmd)

toolbar.add_separator
plugin_menu.add_separator

cmd = ::UI::Command.new('AMS CW Toggle Draped') {
  model = ::Sketchup.active_model
  next unless model
  definitions = {}
  pins = []
  model.selection.each { |e|
    next if !e.is_a?(::Sketchup::Group) && !e.is_a?(::Sketchup::ComponentInstance)
    type = e.get_attribute(cw_gdict, 'Type', nil)
    if type == 1
      definition = AMS::ClothWorks.get_definition(e)
      definitions[definition] = definition.entities
    elsif type == 2
      pins << e
    end
  }
  rcount = 0
  ucount = 0
  definitions.each { |d, s|
    urs = d.get_attribute(cw_gdict, 'State', nil)
    if urs == 0
      rcount += 1
    elsif urs == 1
      ucount += 1
    end
  }
  pins.each { |e|
    if e.get_attribute(cw_gdict, 'Pin Pos N')
      rcount += 1
    elsif e.get_attribute(cw_gdict, 'Pin Pos O')
      ucount += 1
    end
  }
  if rcount > ucount
    op = 'CW - Drape'
    model.start_operation(op, true)
    begin
      definitions.each { |definition, entities|
        if definition.get_attribute(cw_gdict, 'State', nil) == 0 && AMS::ClothWorks.validate_cloth_memory(definition)
          mesh_data = AMS::ClothWorks.compute_mesh_uvqs(entities)
          AMS::ClothWorks.swap_drape_cloth(definition)
          AMS::ClothWorks.position_texture(mesh_data)
          if AMS::ClothWorks::SU_MAJOR_VERSION > 6
            definition.invalidate_bounds
          end
        end
      }
      pins.each { |e|
        new_pos = e.get_attribute(cw_gdict, 'Pin Pos N')
        if new_pos.is_a?(Array)
          e.set_attribute(cw_gdict, 'Pin Pos O', e.transformation.to_a)
          e.transformation = ::Geom::Transformation.new(new_pos)
        end
        e.delete_attribute(cw_gdict, 'Pin Pos N')
      }
    rescue Exception => err
      model.abort_operation
      raise err
    end
    model.commit_operation
  else
    op = 'CW - Undrape'
    model.start_operation(op, true)
    begin
      definitions.each { |definition, entities|
        if definition.get_attribute(cw_gdict, 'State', nil) == 1 && AMS::ClothWorks.validate_cloth_memory(definition)
          mesh_data = AMS::ClothWorks.compute_mesh_uvqs(entities)
          AMS::ClothWorks.swap_drape_cloth(definition)
          AMS::ClothWorks.position_texture(mesh_data)
          if AMS::ClothWorks::SU_MAJOR_VERSION > 6
            definition.invalidate_bounds
          end
        end
      }
      pins.each { |e|
        orig_pos = e.get_attribute(cw_gdict, 'Pin Pos O')
        if orig_pos.is_a?(Array)
          e.set_attribute(cw_gdict, 'Pin Pos N', e.transformation.to_a)
          e.transformation = ::Geom::Transformation.new(orig_pos)
        end
        e.delete_attribute(cw_gdict, 'Pin Pos O')
      }
    rescue Exception => err
      model.abort_operation
      raise err
    end
    model.commit_operation
  end
}
cmd.set_validation_proc {
  next MF_ENABLED unless AMS::ClothWorks::IS_PLATFORM_WINDOWS
  model = ::Sketchup.active_model
  next MF_GRAYED unless model
  next MF_GRAYED if model.active_entities != model.entities
  next MF_GRAYED if AMS::ClothWorks.simulation_active?
  definitions = {}
  pins = []
  model.selection.each { |e|
    next if !e.is_a?(::Sketchup::Group) && !e.is_a?(::Sketchup::ComponentInstance)
    type = e.get_attribute(cw_gdict, 'Type', nil)
    if type == 1
      definition = AMS::ClothWorks.get_definition(e)
      definitions[definition] = true
    elsif type == 2
      pins << e
    end
  }
  rcount = 0
  ucount = 0
  definitions.each { |d, s|
    urs = d.get_attribute(cw_gdict, 'State', nil)
    if urs == 0
      rcount += 1
    elsif urs == 1
      ucount += 1
    end
  }
  pins.each { |e|
    if e.get_attribute(cw_gdict, 'Pin Pos N')
      rcount += 1
    elsif e.get_attribute(cw_gdict, 'Pin Pos O')
      ucount += 1
    end
  }
  next MF_GRAYED if rcount + ucount == 0
  ucount > rcount ? MF_CHECKED : MF_UNCHECKED
}
cmd.menu_text = cmd.tooltip = translate['Toggle Drape']
cmd.status_bar_text = translate['Toggle drape the selected cloth(s) and/or pin(s).']
cmd.small_icon = simg_path + 'toggle_draped.png'
cmd.large_icon = limg_path + 'toggle_draped.png'
toolbar.add_item(cmd)
plugin_menu.add_item(cmd)

pin_tool = nil
cmd = ::UI::Command.new('AMS CW Pin Tool') {
  model = ::Sketchup.active_model
  pin_tool = AMS::ClothWorks::PinTool.new('pin', model)
  model.select_tool(pin_tool)
}
cmd.set_validation_proc {
  next MF_GRAYED unless ::Sketchup.active_model
  pin_tool && pin_tool.active? ? MF_CHECKED : MF_UNCHECKED
}
cmd.menu_text = cmd.tooltip = translate['Add Pin']
cmd.status_bar_text = translate["Attach a pin at a specific location on cloth."]
cmd.small_icon = simg_path + 'pin.png'
cmd.large_icon = limg_path + 'pin.png'

toolbar.add_item(cmd)
plugin_menu.add_item(cmd)

# Extension menu

cmd = ::UI::Command.new('AMS CW Scale Pins') {
  model = ::Sketchup.active_model
  scale = AMS::ClothWorks.get_pin_scale
  prompts = [translate['Scale']]
  defaults = [sprintf("%0.2f", scale)]
  list = ['0.05|0.10|0.25|0.50|0.75|1.00|1.25|1.50|1.75|2.00|3.00|4.00|5.00|10.00']
  input = ::UI.inputbox(prompts, defaults, list, translate['Edit Pins Scale'])
  next unless input
  scale = input[0].to_f
  op = 'CW - Scale Pins'
  model.start_operation(op, true)
  begin
    count = AMS::ClothWorks.set_all_pins_scale(scale)
    feedback = sprintf(translate["Edited the scale of %d pin(s)."], count)
    ::UI.messagebox(feedback)
  rescue Exception => err
    model.abort_operation
    raise err
  end
  model.commit_operation
}
cmd.set_validation_proc {
  ::Sketchup.active_model ? MF_ENABLED : MF_GRAYED
}
cmd.menu_text = translate['Edit Pins Scale']
cmd.status_bar_text = translate['Change the scale of all pins.']
plugin_menu.add_item(cmd)


plugin_menu.add_separator


title = translate['Export Recorded to Images']
cmd = ::UI::Command.new('AMS CW ' + title) {
  if AMS::ClothWorks.record_instance
    AMS::ClothWorks.record_instance.export_to_images
  end
}
cmd.set_validation_proc {
  next MF_GRAYED unless ::Sketchup.active_model
  next MF_GRAYED if AMS::ClothWorks.record_instance.nil? || AMS::ClothWorks.record_instance.data_empty?
  next MF_GRAYED if AMS::ClothWorks.simulation_active?
  MF_ENABLED
}
cmd.menu_text = title
cmd.status_bar_text = translate["Export the simulation record into a sequence of BMP, JPG, PNG, or TIF images."]
plugin_menu.add_item(cmd)

=begin
title = translate['Export Recorded to Video']
cmd = ::UI::Command.new('AMS CW ' + title) {
  if AMS::ClothWorks.record_instance
    AMS::ClothWorks.record_instance.export_to_video
  end
}
cmd.set_validation_proc {
  next MF_GRAYED unless ::Sketchup.active_model
  next MF_GRAYED if AMS::ClothWorks.record_instance.nil? || AMS::ClothWorks.record_instance.data_empty?
  next MF_GRAYED if AMS::ClothWorks.simulation_active?
  MF_ENABLED
}
cmd.menu_text = title
cmd.status_bar_text = translate["Export the simulation record into MP4, MOV, AVI, or other video format. This requires FFmpeg!"]
plugin_menu.add_item(cmd)
=end

title = translate['Export Recorded to SKP']
cmd = ::UI::Command.new('AMS CW ' + title) {
  if AMS::ClothWorks.record_instance
    AMS::ClothWorks.record_instance.export_to_skp
  end
}
cmd.set_validation_proc {
  next MF_GRAYED unless ::Sketchup.active_model
  next MF_GRAYED if AMS::ClothWorks.record_instance.nil? || AMS::ClothWorks.record_instance.data_empty?
  next MF_GRAYED if AMS::ClothWorks.simulation_active?
  MF_ENABLED
}
cmd.menu_text = title
cmd.status_bar_text = translate["Export the simulation record into a sequence of SKP files."]
plugin_menu.add_item(cmd)

title = translate['Export Recorded to Other']
cmd = ::UI::Command.new('AMS CW ' + title) {
  if AMS::ClothWorks.record_instance
    AMS::ClothWorks.record_instance.export_to_other
  end
}
cmd.set_validation_proc {
  next MF_GRAYED unless ::Sketchup.active_model
  next MF_GRAYED if AMS::ClothWorks.record_instance.nil? || AMS::ClothWorks.record_instance.data_empty?
  next MF_GRAYED if AMS::ClothWorks.simulation_active?
  MF_ENABLED
}
cmd.menu_text = title
cmd.status_bar_text = translate["Export simulation record into a sequence of KMZ, DAE, 3DS, DWG, DXF, FBX, OBJ, WRL, XSI, or PDF files."]
plugin_menu.add_item(cmd)


title = translate['Dispose Recorded']
cmd = ::UI::Command.new('AMS CW ' + title) {
  AMS::ClothWorks.clear_record
}
cmd.set_validation_proc {
  next MF_GRAYED unless ::Sketchup.active_model
  next MF_GRAYED if AMS::ClothWorks.record_instance.nil? || AMS::ClothWorks.record_instance.data_empty?
  next MF_GRAYED if AMS::ClothWorks.simulation_active?
  MF_ENABLED
}
cmd.menu_text = title
cmd.status_bar_text = translate["Erase the simulation record."]
plugin_menu.add_item(cmd)


plugin_menu.add_separator


title = translate['Clear Draped and Purge']
cmd = ::UI::Command.new('AMS CW ' + title) {
  op = 'CW - Clear and Purge'
  model = ::Sketchup.active_model
  next unless model
  model.start_operation(op, true)
  model.definitions.purge_unused
  begin
    model.definitions.each { |definition|
      bfound = false
      definition.instances.each { |inst|
        if inst.get_attribute(cw_gdict, 'Type') == 1
          bfound = true
          break
        end
      }
      AMS::ClothWorks.clear_cloth_memory(definition) if bfound
    }
  rescue Exception => err
    model.abort_operation
    raise err
  end
  model.commit_operation
}
cmd.menu_text = title
cmd.status_bar_text = translate["Clear drape memory and purge unused component definitions to reduce file size."]
plugin_menu.add_item(cmd)


title = translate['Reset Model Settings']
cmd = ::UI::Command.new('AMS CW ' + title) {
  op = 'CW - Reset Model Options'
  model = ::Sketchup.active_model
  next unless model
  model.start_operation(op, true)
  AMS::ClothWorks.reset_model_options(model)
  model.commit_operation
  AMS::ClothWorks::UI.update_state(false)
}
cmd.set_validation_proc {
  (::Sketchup.active_model && !AMS::ClothWorks.simulation_active?) ? MF_ENABLED : MF_GRAYED
}
cmd.menu_text = title
cmd.status_bar_text = translate["Reset the settings assigned to the model, that is, the parameters of the simulation and fabric materials."]
plugin_menu.add_item(cmd)


title = translate['Reset Global Settings']
cmd = ::UI::Command.new('AMS CW ' + title) {
  AMS::ClothWorks.reset_global_options
  AMS::ClothWorks::Record.reset_global_options
  AMS::ClothWorks::UI.update_state(false)
}
cmd.menu_text = title
cmd.status_bar_text = translate["Reset global settings, that is, the parameters shared among all models."]
plugin_menu.add_item(cmd)


plugin_menu.add_separator


title = translate['Import Materials']
cmd = ::UI::Command.new('AMS CW ' + title) {
  result = ::UI.openpanel(translate['Select a file to import from...'], nil, "TXT|*.txt||")
  if result
    sfile = ::String.new(result)
    sfile.force_encoding('UTF-8') unless AMS::ClothWorks::IS_RUBY_VERSION18
    AMS::ClothWorks.import_shared_materials(::Sketchup.active_model, sfile)
    AMS::ClothWorks::UI.update_state(false)
  end
}
cmd.set_validation_proc {
  ::Sketchup.active_model ? MF_ENABLED : MF_GRAYED
}
cmd.menu_text = title
cmd.status_bar_text = translate["Import cloth materials from a text file."]
plugin_menu.add_item(cmd)


title = translate['Export Materials']
cmd = ::UI::Command.new('AMS CW ' + title) {
  model_fname = ::File.basename(::Sketchup.active_model.path, '.skp')
  model_fname.force_encoding('UTF-8') unless AMS::ClothWorks::IS_RUBY_VERSION18
  model_fname = "cw_materials" if model_fname.empty?
  result = ::UI.savepanel(translate['Choose a file to export to...'], nil, model_fname + '.txt')
  if result
    sfile = ::String.new(result)
    sfile.force_encoding('UTF-8') unless AMS::ClothWorks::IS_RUBY_VERSION18
    fpath = ::File.dirname(sfile)
    fname = ::File.basename(sfile, '.*') + '.txt'
    AMS::ClothWorks.export_shared_materials(::Sketchup.active_model, ::File.join(fpath, fname))
  end
}
cmd.set_validation_proc {
  ::Sketchup.active_model ? MF_ENABLED : MF_GRAYED
}
cmd.menu_text = title
cmd.status_bar_text = translate["Export cloth materials to a text file."]
plugin_menu.add_item(cmd)

=begin
cmd = ::UI::Command.new('AMS CW FFmpeg Manual') {
  dir = File.dirname(__FILE__)
  dir.force_encoding('UTF-8') unless AMS::ClothWorks::IS_RUBY_VERSION18
  ::UI.openURL(File.join(dir, 'FFmpeg Manual.pdf'))
}
cmd.menu_text = translate['FFmpeg Guide']
cmd.status_bar_text = translate["Open ClothWorks FFmpeg manual."]
plugin_menu.add_item(cmd)
=end

plugin_menu.add_separator

lang_menu = plugin_menu.add_submenu(translate['Language'])

cfpath = __FILE__.dup
cfpath.force_encoding('UTF-8') unless AMS::ClothWorks::IS_RUBY_VERSION18
dir = ::File.expand_path(::File.dirname(cfpath))
fname = ::File.join(dir, 'languages.json')
valid_langs = {}
if ::File.exists?(fname)
  begin
    dat = nil
    ::File.open(fname, 'r') { |fp|
      dat = fp.read
    }
    langs = JSON.parse(dat)
    langs.each { |k, v|
      continue if !k.is_a?(String) || !v.is_a?(String)
      valid_langs[k] = v
    }
  rescue Exception => err
    puts err
    puts err.backtrace
  end
end
if !valid_langs.values.include?('en-US')
  valid_langs["English (US)"] = "en-US"
end
valid_lang_values = valid_langs.values
valid_langs.each { |k, v|
  cmd = ::UI::Command.new('AMS CW Language: ' + v) {
    AMS::ClothWorks.language = v
    ::UI.messagebox(translate.sprintf("Restart SketchUp to apply language: %s", k))
  }
  if (v == 'en-US')
    cmd.set_validation_proc {
      AMS::ClothWorks.language == v || !valid_lang_values.include?(AMS::ClothWorks.language) ? MF_CHECKED : MF_UNCHECKED
    }
  else
    cmd.set_validation_proc {
      AMS::ClothWorks.language == v ? MF_CHECKED : MF_UNCHECKED
    }
  end
  cmd.menu_text = k
  cmd.status_bar_text = k
  lang_menu.add_item(cmd)
}

cmd = ::UI::Command.new('AMS CW About') {
  msg = sprintf("%s %s - %s\n%s © %d - %d, %s",
    translate["ClothWorks version"],
    AMS::ClothWorks::VERSION,
    AMS::ClothWorks::RELEASE_DATE,
    translate["Copyright"],
    AMS::ClothWorks::RELEASE_YEAR_START,
    AMS::ClothWorks::RELEASE_YEAR_END,
    translate["Anton Synytsia"])
  ::UI.messagebox(msg)
}
cmd.menu_text = translate['About']
cmd.status_bar_text = translate["Check the version of ClothWorks and additional information."]
plugin_menu.add_item(cmd)


# Create edit menus

::UI.add_context_menu_handler { |cmenu|
  next if AMS::ClothWorks.simulation_active?
  model = ::Sketchup.active_model

  lr = AMS::ClothWorks.licensed?
  ct = AMS::ClothWorks.time
  cwls = true

  if model.active_entities != model.entities
    # Here we would want to add an option to ignore particular sub-colliders
    tg = model.active_path.first
    type = tg.get_attribute(cw_gdict, 'Type', nil)

    if type == 0
      sub_colliders = []
      model.selection.each { |e|
        if e.is_a?(::Sketchup::Group) ||e.is_a?(::Sketchup::ComponentInstance)
          sub_colliders << e
        end
      }
      unless sub_colliders.empty?
        cw_cmenu = cmenu.add_submenu(translate['ClothWorks'])
        sscol = sub_colliders.size > 1 ? "Sub-Colliders" : "Sub-Collider"
        smitem = sprintf("%d %s", sub_colliders.size, translate[sscol])
        collider_menu = cw_cmenu.add_submenu(smitem)
        if sub_colliders.size > 1
          collider_menu.add_item(translate['Include in Simulation']) {
            op = 'CW - Include in Simulation'
            model.start_operation(op, true)
            sub_colliders.each { |e|
              e.delete_attribute(cw_gdict, 'Exclude')
            }
            model.commit_operation
          }
          collider_menu.add_item(translate['Exclude from Simulation']) {
            op = 'CW - Exclude from Simulation'
            model.start_operation(op, true)
            sub_colliders.each { |e|
              e.set_attribute(cw_gdict, 'Exclude', true)
            }
            model.commit_operation
          }
        else
          sub_collider = sub_colliders.first
          if sub_collider.get_attribute(cw_gdict, 'Exclude', false)
            collider_menu.add_item(translate['Include in Simulation']) {
              op = 'CW - Include in Simulation'
              model.start_operation(op, true)
              sub_collider.delete_attribute(cw_gdict, 'Exclude')
              model.commit_operation
            }
          else
            collider_menu.add_item(translate['Exclude from Simulation']) {
              op = 'CW - Exclude from Simulation'
              model.start_operation(op, true)
              sub_collider.set_attribute(cw_gdict, 'Exclude', true)
              model.commit_operation
            }
          end
        end
      end
    end
    next
  end

  colliders_count = 0
  cloths_count = 0
  pins_count = 0
  voids_count = 0
  unset_count = 0
  all_aggregates = {}
  model.selection.each { |e|
    next if !e.is_a?(::Sketchup::Group) && !e.is_a?(::Sketchup::ComponentInstance)
    type = e.get_attribute(cw_gdict, 'Type', nil)
    if type == 0
      colliders_count += 1
      all_aggregates[e] = 0
    elsif type == 1
      cloths_count += 1
      all_aggregates[e] = 1
    elsif type == 2
      pins_count += 1
      all_aggregates[e] = 2
    elsif type == 3
      voids_count += 1
      all_aggregates[e] = 3
    else
      unset_count += 1
      all_aggregates[e] = -1
    end
  }
  next if all_aggregates.empty?
  cw_cmenu = cmenu.add_submenu('ClothWorks')

  count = colliders_count + pins_count + voids_count + unset_count
  if count > 0
    cw_cmenu.add_item(translate["Make Cloth#{count > 1 ? 's' : ''}"]) {
      op = 'CW - Make Cloth'
      model.start_operation(op, true)
      all_aggregates.each { |e, s|
        e.set_attribute(cw_gdict, 'Type', 1)
      }
      model.commit_operation
      AMS::ClothWorks::UI.update_state(true)
    }
  end

  count = cloths_count + pins_count + voids_count + unset_count
  if count > 0
    cw_cmenu.add_item(translate["Make Collider#{count > 1 ? 's' : ''}"]) {
      op = 'CW - Make Collider'
      model.start_operation(op, true)
      all_aggregates.each { |e, s|
        e.set_attribute(cw_gdict, 'Type', 0)
      }
      model.commit_operation
      AMS::ClothWorks::UI.update_state(true)
    }
  end

  count = cloths_count + colliders_count + voids_count + unset_count
  if count > 0
    cw_cmenu.add_item(translate["Make Pin#{count > 1 ? 's' : ''}"]) {
      op = 'CW - Make Pin'
      model.start_operation(op, true)
      all_aggregates.each { |e, s|
        e.set_attribute(cw_gdict, 'Type', 2)
      }
      model.commit_operation
      AMS::ClothWorks::UI.update_state(true)
    }
  end

  count = cloths_count + colliders_count + pins_count + unset_count
  if count > 0
    cw_cmenu.add_item(translate["Make Void#{count > 1 ? 's' : ''}"]) {
      op = 'CW - Make Void'
      model.start_operation(op, true)
      all_aggregates.each { |e, s|
        e.set_attribute(cw_gdict, 'Type', 3)
      }
      model.commit_operation
      AMS::ClothWorks::UI.update_state(true)
    }
  end

  count = colliders_count + cloths_count + pins_count + voids_count
  if count > 0
    cw_cmenu.add_item(translate["Ignore#{count > 1 ? ' All' : ''}"]) {
      op = 'CW - Ignore'
      model.start_operation(op, true)
      all_aggregates.each { |e, s|
        e.delete_attribute(cw_gdict, 'Type')
      }
      model.commit_operation
      AMS::ClothWorks::UI.update_state(true)
    }
  end

  separator_added = false

  if cloths_count > 0
    cw_cmenu.add_separator
    separator_added = true

    sscloth = cloths_count > 1 ? "Cloths" : "Cloth"
    smitem = sprintf("%d %s", cloths_count, translate[sscloth])
    cloth_menu = cw_cmenu.add_submenu(smitem)


if cwls

    cloth_menu.add_item(translate['Apply Adaptive Grid']) {
      # Prompt
      prompts = [translate['Padding']]
      defaults = [AMS::ClothWorks.grid_padding.to_l.to_s]
      list = ['']

      val = nil
      while true
        results = ::UI.inputbox(prompts, defaults, list, translate['Grid Parameters'])
        break unless results
        begin
          val = results[0].to_l
          break
        rescue Exception => err
          msg = sprintf("%s\n\n%s", err.message, translate["Please try again."])
          ::UI.messagebox(msg)
        end
      end
      next unless val
      AMS::ClothWorks.grid_padding = val

      # Filter selection
      instances = []
      all_aggregates.each { |e, s|
        next if s != 1
        instances << e
      }

      # Check for stray, soft/smooth geometry
      if AMS::ClothWorks.multiple_contain_stray_geometry?(instances, true) && ::UI.messagebox(translate["Purge the selected cloth of soft and smooth, coplanar geometry, prior to generating a grid?"], MB_YESNO) == IDYES
        model.start_operation('CW - Purge Geometry', true)

        begin
          AMS::ClothWorks.purge_edges_multiple(instances, true)

        rescue Exception => err
          model.abort_operation
          raise err
        end

        model.commit_operation
      end

      # Apply grid
      model.start_operation('CW - Apply Adaptive Grid', true)

      begin
        # Make groups unique
        instances.reject! { |e|
          next true unless e.valid?
          e.make_unique if e.is_a?(::Sketchup::Group)
          false
        }

        # Generate Grid
        ci, err_codes = AMS::ClothWorks.apply_adaptive_grid_multiple(instances, AMS::ClothWorks.grid_padding, AMS::ClothWorks::MESH_MANIPULATION_VERTEX_LIMIT)

        # Replace original instances with new ones
        ci.each { |oe, ne|
          oe.erase! if oe.valid?

          model.selection.add(ne)
        }

      rescue Exception => err
        model.abort_operation
        raise err
      end

      model.commit_operation

      op = 'CW - Purge Definitions'
      model.start_operation(op, true)
      model.definitions.purge_unused
      model.commit_operation
    }

else

    cloth_menu.add_item(translate['Apply Adaptive Grid']) {
      msg = sprintf("%s\n%s.\n\n%s\n(%s) %s → %s → %s",
        translate["Please license ClothWorks to unlock"],
        translate["Adaptive Grid"],
        translate["To license, access:"],
        translate["Menu"],
        translate["Extensions"],
        translate["ClothWorks"],
        translate["License..."])
      ::UI.messagebox(msg)
    }

end

    cloth_menu.add_item(translate['Apply Quadrilateral Grid']) {
      # Prompt
      prompts = [translate['Resolution']]
      defaults = [AMS::ClothWorks.grid_resolution.to_s]
      list = ['']

      val = nil
      while true
        results = ::UI.inputbox(prompts, defaults, list, translate['Grid Parameters'])
        break unless results
        begin
          val = results[0].to_i
          break
        rescue Exception => err
          msg = sprintf("%s\n\n%s", err.message, translate["Please try again."])
          ::UI.messagebox(msg)
        end
      end
      next unless val
      AMS::ClothWorks.grid_resolution = val

      # Filter selection
      instances = []
      all_aggregates.each { |e, s|
        next if s != 1
        instances << e
      }

      # Check for stray geometry
      if AMS::ClothWorks.multiple_contain_stray_geometry?(instances, false) && ::UI.messagebox(translate["Purge the selected cloth of coplanar geometry, prior to generating a grid?"], MB_YESNO) == IDYES
        model.start_operation('CW - Purge Geometry', true)

        begin
          AMS::ClothWorks.purge_edges_multiple(instances, false)

        rescue Exception => err
          model.abort_operation
          raise err
        end

        model.commit_operation
      end

      # Apply grid
      model.start_operation('CW - Quadrilateral Grid', true)

      begin
        # Make groups unique
        instances.reject! { |e|
          next true unless e.valid?
          e.make_unique if e.is_a?(::Sketchup::Group)
          false
        }

        # Generate Grid
        ci, err_codes = AMS::ClothWorks.apply_quadrilateral_grid_multiple(instances, AMS::ClothWorks.grid_resolution, AMS::ClothWorks::MESH_MANIPULATION_VERTEX_LIMIT)

        # Replace original instances with new ones
        ci.each { |oe, ne|
          oe.erase! if oe.valid?

          model.selection.add(ne)
        }

      rescue Exception => err
        model.abort_operation
        raise err
      end

      model.commit_operation

      # Purge unused
      op = 'CW - Purge Definitions'
      model.start_operation(op, true)
      model.definitions.purge_unused
      model.commit_operation

      # Display errors
      err_codes2 = {}
      err_codes.each { |d, eid|
        if err_codes2.has_key?(eid)
          err_codes2[eid] += 1
        else
          err_codes2[eid] = 1
        end
      }

      if err_codes2[1] || err_codes2[2] || err_codes2[3]
        count = 0
        count += err_codes2[1] if err_codes2[1]
        count += err_codes2[2] if err_codes2[2]
        count += err_codes2[3] if err_codes2[3]
        msg = translate.sprintf("Failed to generate a quadrilateral grid for %d component definition(s) due to one or more of the following reason(s):", count)
        msg << "\n\n"
        if err_codes2[1]
          msg << "• "
          msg << translate["Initial topology conditions for applying a grid are not met. Initial mesh geometry should consist of one quadrilateral face or two triplets, linked by a diagonal edge."]
          msg << "\n\n"
        end
        if err_codes2[2]
          msg << "• "
          msg << translate["Face area is too small or grid spacing is too large. Increase face size or decrease grid spacing (increase grid resolution) to resolve the issue."]
          msg << "\n\n"
        end
        if err_codes2[3]
          msg << "• "
          msg << translate["Grid resolution exceeds the vertex count limit."]
          msg << " "
          msg << translate.sprintf("The vertex count limit is %d vertices per every cloth mesh.", AMS::ClothWorks::MESH_MANIPULATION_VERTEX_LIMIT)
          msg << " "
          msg << translate["Vertex count limit is imposed to prevent significant delays when generating a grid."]
          msg << " "
          msg << translate["Increase grid spacing (reduce grid resolution) to resolve the issue."]
          msg << "\n\n"
        end
        ::UI.messagebox(msg)
      end
    }

    cloth_menu.add_item(translate['Apply Cross-Quadrilateral Grid']) {
      # Prompt
      prompts = [translate['Resolution']]
      defaults = [AMS::ClothWorks.grid_resolution.to_s]
      list = ['']

      val = nil
      while true
        results = ::UI.inputbox(prompts, defaults, list, translate['Grid Parameters'])
        break unless results
        begin
          val = results[0].to_i
          break
        rescue Exception => err
          msg = sprintf("%s\n\n%s", err.message, translate["Please try again."])
          ::UI.messagebox(msg)
        end
      end
      next unless val
      AMS::ClothWorks.grid_resolution = val

      # Filter selection
      instances = []
      all_aggregates.each { |e, s|
        next if s != 1
        instances << e
      }

      # Check for stray geometry
      if AMS::ClothWorks.multiple_contain_stray_geometry?(instances, false) && ::UI.messagebox(translate["Purge the selected cloth of coplanar geometry, prior to generating a grid?"], MB_YESNO) == IDYES
        model.start_operation('CW - Purge Geometry', true)

        begin
          AMS::ClothWorks.purge_edges_multiple(instances, false)

        rescue Exception => err
          model.abort_operation
          raise err
        end

        model.commit_operation
      end

      # Apply grid
      model.start_operation('CW - Cross-Quadrilateral Grid', true)

      begin
        # Make groups unique
        instances.reject! { |e|
          next true unless e.valid?
          e.make_unique if e.is_a?(::Sketchup::Group)
          false
        }

        # Generate Grid
        ci, err_codes = AMS::ClothWorks.apply_cross_quadrilateral_grid_multiple(instances, AMS::ClothWorks.grid_resolution, AMS::ClothWorks::MESH_MANIPULATION_VERTEX_LIMIT)

        # Replace original instances with new ones
        ci.each { |oe, ne|
          oe.erase! if oe.valid?

          model.selection.add(ne)
        }

      rescue Exception => err
        model.abort_operation
        raise err
      end

      model.commit_operation

      # Purge unused
      op = 'CW - Purge Definitions'
      model.start_operation(op, true)
      model.definitions.purge_unused
      model.commit_operation

      # Display errors
      err_codes2 = {}
      err_codes.each { |d, eid|
        if err_codes2.has_key?(eid)
          err_codes2[eid] += 1
        else
          err_codes2[eid] = 1
        end
      }

      if err_codes2[1] || err_codes2[2] || err_codes2[3]
        count = 0
        count += err_codes2[1] if err_codes2[1]
        count += err_codes2[2] if err_codes2[2]
        count += err_codes2[3] if err_codes2[3]
        msg = translate.sprintf("Failed to generate a quadrilateral grid for %d component definition(s) due to one or more of the following reason(s):", count)
        msg << "\n\n"
        if err_codes2[1]
          msg << "• "
          msg << translate["Initial topology conditions for applying a grid are not met. Initial mesh geometry should consist of one quadrilateral face or two triplets, linked by a diagonal edge."]
          msg << "\n\n"
        end
        if err_codes2[2]
          msg << "• "
          msg << translate["Face area is too small or grid spacing is too large. Increase face size or decrease grid spacing (increase grid resolution) to resolve the issue."]
          msg << "\n\n"
        end
        if err_codes2[3]
          msg << "• "
          msg << translate["Grid resolution exceeds the vertex count limit."]
          msg << " "
          msg << translate.sprintf("The vertex count limit is %d vertices per every cloth mesh.", AMS::ClothWorks::MESH_MANIPULATION_VERTEX_LIMIT)
          msg << " "
          msg << translate["Vertex count limit is imposed to prevent significant delays when generating a grid."]
          msg << " "
          msg << translate["Increase grid spacing (reduce grid resolution) to resolve the issue."]
          msg << "\n\n"
        end
        ::UI.messagebox(msg)
      end
    }

    cloth_menu.add_item(translate['Apply Equilateral Grid']) {
      # Prompt
      prompts = [translate['Resolution']]
      defaults = [AMS::ClothWorks.grid_resolution.to_s]
      list = ['']

      val = nil
      while true
        results = ::UI.inputbox(prompts, defaults, list, translate['Grid Parameters'])
        break unless results
        begin
          val = results[0].to_i
          break
        rescue Exception => err
          msg = sprintf("%s\n\n%s", err.message, translate["Please try again."])
          ::UI.messagebox(msg)
        end
      end
      next unless val
      AMS::ClothWorks.grid_resolution = val

      # Filter selection
      instances = []
      all_aggregates.each { |e, s|
        next if s != 1
        instances << e
      }

      # Check for stray geometry
      if AMS::ClothWorks.multiple_contain_stray_geometry?(instances, false) && ::UI.messagebox(translate["Purge the selected cloth of coplanar geometry, prior to generating a grid?"], MB_YESNO) == IDYES
        model.start_operation('CW - Purge Geometry', true)

        begin
          AMS::ClothWorks.purge_edges_multiple(instances, false)

        rescue Exception => err
          model.abort_operation
          raise err
        end

        model.commit_operation
      end

      # Apply grid
      model.start_operation('CW - Equilateral Grid', true)

      begin
        # Make groups unique
        instances.reject! { |e|
          next true unless e.valid?
          e.make_unique if e.is_a?(::Sketchup::Group)
          false
        }

        # Generate Grid
        ci, err_codes = AMS::ClothWorks.apply_equilateral_grid_multiple(instances, AMS::ClothWorks.grid_resolution, AMS::ClothWorks::MESH_MANIPULATION_VERTEX_LIMIT)

        # Replace original instances with new ones
        ci.each { |oe, ne|
          oe.erase! if oe.valid?

          model.selection.add(ne)
        }

      rescue Exception => err
        model.abort_operation
        raise err
      end

      model.commit_operation

      # Purge
      op = 'CW - Purge Definitions'
      model.start_operation(op, true)
      model.definitions.purge_unused
      model.commit_operation

      # Display errors
      err_codes2 = {}
      err_codes.each { |d, eid|
        if err_codes2.has_key?(eid)
          err_codes2[eid] += 1
        else
          err_codes2[eid] = 1
        end
      }

      if err_codes2[1] || err_codes2[2] || err_codes2[3]
        count = 0
        count += err_codes2[1] if err_codes2[1]
        count += err_codes2[2] if err_codes2[2]
        count += err_codes2[3] if err_codes2[3]
        msg = translate.sprintf("Failed to generate an equilateral grid for %d component definition(s) due to one or more of the following reason(s):", count)
        msg << "\n\n"
        if err_codes2[1]
          msg << "• "
          msg << translate["Initial topology conditions for applying a grid are not met. Initial mesh geometry should consist of one quadrilateral face or two triplets, linked by a diagonal edge."]
          msg << "\n\n"
        end
        if err_codes2[2]
          msg << "• "
          msg << translate["Face area is too small or grid spacing is too large. Increase face size or decrease grid spacing (increase grid resolution) to resolve the issue."]
          msg << "\n\n"
        end
        if err_codes2[3]
          msg << "• "
          msg << translate["Grid resolution exceeds the vertex count limit."]
          msg << " "
          msg << translate.sprintf("The vertex count limit is %d vertices per every cloth mesh.", AMS::ClothWorks::MESH_MANIPULATION_VERTEX_LIMIT)
          msg << " "
          msg << translate["Vertex count limit is imposed to prevent significant delays when generating a grid."]
          msg << " "
          msg << translate["Increase grid spacing (reduce grid resolution) to resolve the issue."]
          msg << "\n\n"
        end
        ::UI.messagebox(msg)
      end
    }

    cloth_menu.add_separator

if cwls

    cloth_menu.add_item(translate['Apply Loop Subdivision']) {
      # Filter selection
      instances = []
      all_aggregates.each { |e, s|
        next if s != 1
        instances << e
      }

      # Wrap operation
      model.start_operation('CW - Loop Subdivision', true)

      begin
        # Make groups unique
        instances.reject! { |e|
          next true unless e.valid?
          e.make_unique if e.is_a?(::Sketchup::Group)
          false
        }

        # Generate Grid
        ci, err_codes = AMS::ClothWorks.apply_loop_subdivision_multiple(instances, AMS::ClothWorks::MESH_MANIPULATION_VERTEX_LIMIT)

        # Replace original instances with new ones
        ci.each { |oe, ne|
          oe.erase! if oe.valid?

          # turn off adaptive remeshing
          ne.delete_attribute(cw_gdict, 'Quality')

          # add to selection
          model.selection.add(ne)
        }

      rescue Exception => err
        model.abort_operation
        raise err
      end

      model.commit_operation

      # Purge unused
      op = 'CW - Purge Definitions'
      model.start_operation(op, true)
      model.definitions.purge_unused
      model.commit_operation

      # Display errors
      count = 0
      err_codes.each { |d, s|
        count += 1 unless s
      }

      if count > 0
        msg = translate.sprintf("Failed to apply loop subdivision for %d component definition(s) because vertex count exceeds the limit.", count)
        msg << " "
        msg << translate.sprintf("The vertex count limit is %d vertices per every cloth mesh.", AMS::ClothWorks::MESH_MANIPULATION_VERTEX_LIMIT)
        msg << " "
        msg << translate["Vertex count limit is imposed to prevent significant delays when generating a grid."]
        msg << "\n\n"
        msg << translate["To apply Loop subdivision, exceeding the vertex count limit, press YES; otherwise, press NO. It is recommended to save the model prior to applying Loop subdivision on the high resolution mesh. Upon saving the model, try again and press YES when prompted."]
        if ::UI.messagebox(msg, MB_YESNO) == IDYES
          # Wrap operation
          model.start_operation('CW - Loop Subdivision', true)

          begin
            # Generate Grid
            ci, err_codes = AMS::ClothWorks.apply_loop_subdivision_multiple(instances, nil)

            # Replace original instances with new ones
            ci.each { |oe, ne|
              oe.erase! if oe.valid?

              model.selection.add(ne)
            }

          rescue Exception => err
            model.abort_operation
            raise err
          end

          model.commit_operation

          # Purge unused
          op = 'CW - Purge Definitions'
          model.start_operation(op, true)
          model.definitions.purge_unused
          model.commit_operation
        end
      end
    }

else

    cloth_menu.add_item(translate['Apply Loop Subdivision']) {
      msg = sprintf("%s\n%s.\n\n%s\n(%s) %s → %s → %s",
        translate["Please license ClothWorks to unlock"],
        translate["Loop Subdivision"],
        translate["To license, access:"],
        translate["Menu"],
        translate["Extensions"],
        translate["ClothWorks"],
        translate["License..."])
      ::UI.messagebox(msg)
    }

end
=begin
    cloth_menu.add_item('Decimate Mesh') {
      # Options
      prompts = ['Min T Curvature', 'Min E Curvature', 'Include Interior Sharp Edges']
      defaults = [
        sprintf("%0.5f", AMS::ClothWorks.decimate_min_tcurvature),
        sprintf("%0.5f", AMS::ClothWorks.decimate_min_ecurvature),
        AMS::ClothWorks.decimate_interior_sharp_edges? ? 'Yes' : 'No' ]
      list = ['', '', 'Yes|No']
      results = ::UI.inputbox(prompts, defaults, list, 'Mesh Decimation Options')
      next unless results
      AMS::ClothWorks.decimate_min_tcurvature = results[0].to_f
      AMS::ClothWorks.decimate_min_ecurvature = results[1].to_f
      AMS::ClothWorks.decimate_interior_sharp_edges = (results[1] == 'Yes')

      # Filter selection
      instances = []
      all_aggregates.each { |e, s|
        next if s != 1
        instances << e
      }

      # Wrap operation
      st = Time.now
      t = Time.now
      model.start_operation('CW - Decimate Mesh', true)
      printf("%.6f sec <-- model.start_operation(true)\n", Time.now - t)

      begin
        # Make groups unique
        instances.reject! { |e|
          next true unless e.valid?
          e.make_unique if e.is_a?(::Sketchup::Group)
          false
        }

        # Decimate mesh
        t = Time.now
        ci, err_codes = AMS::ClothWorks.decimate_mesh_multiple(
          instances,
          AMS::ClothWorks.decimate_min_tcurvature,
          AMS::ClothWorks.decimate_min_ecurvature,
          AMS::ClothWorks.decimate_interior_sharp_edges?)
        printf("%.6f sec <-- total read, decimate, and write\n", Time.now - t)

        # Replace original instances with new ones
        t = Time.now
        ci.each { |oe, ne|
          oe.erase! if oe.valid?

          model.selection.add(ne)
        }
        printf("%.6f sec <-- delete original instances via Sketchup::Drawingelement.#erase!\n", Time.now - t)

      rescue Exception => err
        model.abort_operation
        raise err
      end

      t = Time.now
      model.commit_operation
      printf("%.6f sec <-- model.commit_operation\n", Time.now - t)

      t = Time.now
      op = 'CW - Purge Definitions'
      model.start_operation(op, true)
      model.definitions.purge_unused
      model.commit_operation
      printf("%.6f sec <-- model.definitions.purge_unused\n", Time.now - t)
      printf("%.6f sec <-- TOTAL operation\n", Time.now - st)
    }

=begin
    cloth_menu.add_item('Apply Adaptive Subdivision') {
      # Options
      prompts = ['Min Refine']
      defaults = [sprintf("%0.6f", AMS::ClothWorks.adaptive_subdivision_quality)]
      list = ['']
      results = ::UI.inputbox(prompts, defaults, list, 'Adaptive Subdivision Options')
      next unless results
      AMS::ClothWorks.adaptive_subdivision_quality = results[0]

      # Filter selection
      instances = []
      all_aggregates.each { |e, s|
        next if s != 1
        instances << e
      }

      # Wrap operation
      model.start_operation('CW - Apply Adaptive Subdivision', true)

      begin
        # Make groups unique
        instances.reject! { |e|
          next true unless e.valid?
          e.make_unique if e.is_a?(::Sketchup::Group)
          false
        }

        # Generate Grid
        ci, err_codes = AMS::ClothWorks.apply_adaptive_subdivision_multiple(
            instances,
            AMS::ClothWorks.adaptive_subdivision_quality,
            AMS::ClothWorks::ADAPTIVE_SUBDIVISION_MIN_AREA,
            AMS::ClothWorks::ADAPTIVE_SUBDIVISION_MIN_LENGTH,
            AMS::ClothWorks::MESH_MANIPULATION_VERTEX_LIMIT)

        # Replace original instances with new ones
        ci.each { |oe, ne|
          oe.erase! if oe.valid?

          model.selection.add(ne)
        }

      rescue Exception => err
        model.abort_operation
        raise err
      end

      model.commit_operation

      op = 'CW - Purge Definitions'
      model.start_operation(op, true)
      model.definitions.purge_unused
      model.commit_operation
    }
=end
if cwls

    cloth_menu.add_item(translate['Apply Laplace Smoothing']) {
      # Options
      prompts = [translate['Number of Iterations']]
      defaults = [AMS::ClothWorks.laplace_iterations.to_s]
      list = ['']
      results = ::UI.inputbox(prompts, defaults, list, translate['Laplace Smoothing Parameters'])
      next unless results
      AMS::ClothWorks.laplace_iterations = results[0].to_i

      # Filter selection
      instances = []
      all_aggregates.each { |e, s|
        next if s != 1
        instances << e
      }

      # Wrap operation
      model.start_operation('CW - Laplace Smoothing', true)

      begin
        # Make groups unique
        instances.reject! { |e|
          next true unless e.valid?
          e.make_unique if e.is_a?(::Sketchup::Group)
          false
        }

        # Generate Grid
        ci, err_codes = AMS::ClothWorks.apply_laplace_smoothing_multiple(instances, AMS::ClothWorks.laplace_iterations)
        # Replace original instances with new ones
        ci.each { |oe, ne|
          oe.erase! if oe.valid?

          model.selection.add(ne)
        }

      rescue Exception => err
        model.abort_operation
        raise err
      end

      model.commit_operation
      op = 'CW - Purge Definitions'
      model.start_operation(op, true)
      model.definitions.purge_unused
      model.commit_operation
    }

else

    cloth_menu.add_item(translate['Apply Laplace Smoothing']) {
      msg = sprintf("%s\n%s.\n\n%s\n(%s) %s → %s → %s",
        translate["Please license ClothWorks to unlock"],
        translate["Laplace Smoothing"],
        translate["To license, access:"],
        translate["Menu"],
        translate["Extensions"],
        translate["ClothWorks"],
        translate["License..."])
      ::UI.messagebox(msg)
    }

end

if cwls

    cloth_menu.add_item(translate['Generate Thickness']) {
      # Prompt
      prompts = [
        translate['Thickness'],
        translate['Number of String Segments'],
        translate['Smoothen Boundary Edges']]
      stryes = translate['Yes']
      strno = translate['No']
      defaults = [AMS::ClothWorks.mesh_thickness.to_l.to_s, AMS::ClothWorks.string_num_seg, AMS::ClothWorks.smoothen_boundaries? ? stryes : strno]
      drop_downs = ['', '', stryes + '|' + strno]

      val1 = nil
      val2 = nil
      val3 = nil
      while true
        results = ::UI.inputbox(prompts, defaults, drop_downs, translate['Thickness Parameters'])
        break unless results
        begin
          val1 = results[0].to_l
          val2 = results[1].to_i
          val3 = results[2] == stryes
          break
        rescue Exception => err
          val1 = nil
          val2 = nil
          msg = sprintf("%s\n\n%s", err.message, translate["Please try again."])
          ::UI.messagebox(msg)
        end
      end
      next if val1.nil? || val2.nil?
      AMS::ClothWorks.mesh_thickness = val1
      AMS::ClothWorks.string_num_seg = val2
      AMS::ClothWorks.smoothen_boundaries = val3

      # Filter selection
      instances = []
      all_aggregates.each { |e, s|
        next if s != 1
        instances << e
      }

      # Generate thickness
      model.start_operation('CW - Generate Thickness', true)

      begin
        # Make groups unique
        instances.reject! { |e|
          next true unless e.valid?
          e.make_unique if e.is_a?(::Sketchup::Group)
          false
        }

        # Generate Grid
        ci, err_codes = AMS::ClothWorks.generate_thickness_multiple(instances, AMS::ClothWorks.mesh_thickness, AMS::ClothWorks.string_num_seg, AMS::ClothWorks.smoothen_boundaries?)

        # Replace original instances with new ones
        ci.each { |oe, ne|
          oe.erase! if oe.valid?

          model.selection.add(ne)
        }

      rescue Exception => err
        model.abort_operation
        raise err
      end

      model.commit_operation

      op = 'CW - Purge Definitions'
      model.start_operation(op, true)
      model.definitions.purge_unused
      model.commit_operation
    }

else

    cloth_menu.add_item(translate['Generate Thickness']) {
      msg = sprintf("%s\n%s.\n\n%s\n(%s) %s → %s → %s",
        translate["Please license ClothWorks to unlock"],
        translate["Thickness Generator"],
        translate["To license, access:"],
        translate["Menu"],
        translate["Extensions"],
        translate["ClothWorks"],
        translate["License..."])
      ::UI.messagebox(msg)
    }

end

    cloth_menu.add_separator

    cloth_menu.add_item(translate['Purge All Edges']) {
      op = 'CW - Purge All Edges'
      model.start_operation(op, true)
      definitions = {}
      all_aggregates.each { |entity, state|
        next if state != 1
        definition = AMS::ClothWorks.get_definition(entity)
        definitions[definition] = definition.entities
      }
      begin
        definitions.each { |definition, entities|
          AMS::ClothWorks.purge_edges(entities, false)
          AMS::ClothWorks.clear_cloth_memory(definition)
          if AMS::ClothWorks::SU_MAJOR_VERSION > 6
            definition.invalidate_bounds
          end
        }
      rescue Exception => err
        model.abort_operation
        raise err
      end
      model.commit_operation

      op = 'CW - Purge Definitions'
      model.start_operation(op, true)
      model.definitions.purge_unused
      model.commit_operation
    }

     cloth_menu.add_item(translate['Purge Soft Edges']) {
      op = 'CW - Purge Soft and Smooth Edges'
      model.start_operation(op, true)
      definitions = {}
      all_aggregates.each { |entity, state|
        next if state != 1
        definition = AMS::ClothWorks.get_definition(entity)
        definitions[definition] = definition.entities
      }
      begin
        definitions.each { |definition, entities|
          AMS::ClothWorks.purge_edges(entities, true)
          AMS::ClothWorks.clear_cloth_memory(definition)
          if AMS::ClothWorks::SU_MAJOR_VERSION > 6
            definition.invalidate_bounds
          end
        }
      rescue Exception => err
        model.abort_operation
        raise err
      end
      model.commit_operation

      op = 'CW - Purge Definitions'
      model.start_operation(op, true)
      model.definitions.purge_unused
      model.commit_operation
    }

    cloth_menu.add_item(translate['Split Edges']) {
      prompts = [translate['String Padding']]
      defaults = [AMS::ClothWorks.grid_padding.to_l.to_s]
      list = ['']

      val = nil
      while true
        results = ::UI.inputbox(prompts, defaults, list, translate['Split Parameters'])
        break unless results
        begin
          val = results[0].to_l
          break
        rescue Exception => err
          msg = sprintf("%s\n\n%s", err.message, translate["Please try again."])
          ::UI.messagebox(msg)
        end
      end
      next unless val
      AMS::ClothWorks.grid_padding = val

      op = 'CW - Split Edges'
      model.start_operation(op, true)
      definitions = {}
      all_aggregates.each { |entity, state|
        next if state != 1
        definition = AMS::ClothWorks.get_definition(entity)
        definitions[definition] = definition.entities
      }
      begin
        definitions.each { |definition, entities|
          AMS::ClothWorks.clear_cloth_memory(definition)
          AMS::ClothWorks.explode_curve(entities)
          AMS::ClothWorks.split_edges(entities, AMS::ClothWorks.grid_padding)
        }
      rescue Exception => err
        model.abort_operation
        raise err
      end
      model.commit_operation

      op = 'CW - Purge Definitions'
      model.start_operation(op, true)
      model.definitions.purge_unused
      model.commit_operation
    }

    cloth_menu.add_item(translate['Weld Edges']) {
      op = 'CW - Weld Edges'
      model.start_operation(op, true)
      definitions = {}
      all_aggregates.each { |entity, state|
        next if state != 1
        definition = AMS::ClothWorks.get_definition(entity)
        definitions[definition] = definition.entities
      }
      begin
        definitions.each { |definition, entities|
          AMS::ClothWorks.weld_edges(entities)
          AMS::ClothWorks.clear_cloth_memory(definition)
        }
      rescue Exception => err
        model.abort_operation
        raise err
      end
      model.commit_operation

      op = 'CW - Purge Definitions'
      model.start_operation(op, true)
      model.definitions.purge_unused
      model.commit_operation
    }


    cloth_menu.add_separator


    item = cloth_menu.add_item(translate['Undrape']) {
      op = 'CW - Undrape'
      model.start_operation(op, true)
      definitions = {}
      all_aggregates.each { |entity, state|
        next if state != 1
        definition = AMS::ClothWorks.get_definition(entity)
        definitions[definition] = definition.entities
      }
      begin
        definitions.each { |definition, entities|
          if definition.get_attribute(cw_gdict, 'State', nil) == 1 && AMS::ClothWorks.validate_cloth_memory(definition)
            mesh_data = AMS::ClothWorks.compute_mesh_uvqs(entities)
            AMS::ClothWorks.swap_drape_cloth(definition)
            AMS::ClothWorks.position_texture(mesh_data)
            if AMS::ClothWorks::SU_MAJOR_VERSION > 6
              definition.invalidate_bounds
            end
          end
        }
      rescue Exception => err
        model.abort_operation
        raise err
      end
      model.commit_operation
    }
    cloth_menu.set_validation_proc(item) {
      found = false
      all_aggregates.each { |entity, state|
        next if state != 1
        definition = AMS::ClothWorks.get_definition(entity)
        if definition.get_attribute(cw_gdict, 'State', nil) == 1
          found = true
          break
        end
      }
      found ? MF_ENABLED : MF_GRAYED
    }

    item = cloth_menu.add_item(translate['Drape']) {
      op = 'CW - Drape'
      model.start_operation(op, true)
      definitions = {}
      all_aggregates.each { |entity, state|
        next if state != 1
        definition = AMS::ClothWorks.get_definition(entity)
        definitions[definition] = definition.entities
      }
      begin
        definitions.each { |definition, entities|
          if definition.get_attribute(cw_gdict, 'State', nil) == 0 && AMS::ClothWorks.validate_cloth_memory(definition)
            mesh_data = AMS::ClothWorks.compute_mesh_uvqs(entities)
            AMS::ClothWorks.swap_drape_cloth(definition)
            AMS::ClothWorks.position_texture(mesh_data)
            if AMS::ClothWorks::SU_MAJOR_VERSION > 6
              definition.invalidate_bounds
            end
          end
        }
      rescue Exception => err
        model.abort_operation
        raise err
      end
      model.commit_operation
    }
    cloth_menu.set_validation_proc(item) {
      found = false
      all_aggregates.each { |entity, state|
        next if state != 1
        definition = AMS::ClothWorks.get_definition(entity)
        if definition.get_attribute(cw_gdict, 'State', nil) == 0
          found = true
          break
        end
      }
      found ? MF_ENABLED : MF_GRAYED
    }

    item = cloth_menu.add_item(translate['Clear Memory']) {
      op = 'CW - Clear Memory'
      model.start_operation(op, true)
      definitions = {}
      all_aggregates.each { |entity, state|
        next if state != 1
        definition = AMS::ClothWorks.get_definition(entity)
        definitions[definition] = entity.transformation
      }
      begin
        definitions.each { |definition, transformation|
          if definition.get_attribute(cw_gdict, 'State', nil)
            AMS::ClothWorks.clear_cloth_memory(definition)
          end
        }
      rescue Exception => err
        model.abort_operation
        raise err
      end
      model.commit_operation
    }
    cloth_menu.set_validation_proc(item) {
      found = false
      all_aggregates.each { |entity, state|
        next if state != 1
        definition = AMS::ClothWorks.get_definition(entity)
        if definition.get_attribute(cw_gdict, 'State', nil)
          found = true
          break
        end
      }
      found ? MF_ENABLED : MF_GRAYED
    }

    cloth_menu.add_separator

    cloth_menu.add_item(translate['Ignore']) {
      op = 'CW - Ignore'
      model.start_operation(op, true)
      all_aggregates.each { |e, s|
        next if s != 1
        e.delete_attribute(cw_gdict, 'Type')
      }
      model.commit_operation
      AMS::ClothWorks::UI.update_state(true)
    }
  end

  if colliders_count > 0
    unless separator_added
      cw_cmenu.add_separator
      separator_added = true
    end
    sscol = colliders_count > 1 ? "Colliders" : "Collider"
    sitem = sprintf("%d %s", colliders_count, translate[sscol])
    collider_menu = cw_cmenu.add_submenu(sitem)

    collider_menu.add_item(translate['Ignore']) {
      op = 'CW - Ignore'
      model.start_operation(op, true)
      all_aggregates.each { |e, s|
        next if s != 0
        e.delete_attribute(cw_gdict, 'Type')
      }
      model.commit_operation
      AMS::ClothWorks::UI.update_state(true)
    }
  end

  if pins_count > 0
    unless separator_added
      cw_cmenu.add_separator
      separator_added = true
    end

    sspin = pins_count > 1 ? "Pins" : "Pin"
    sitem = sprintf("%d %s", pins_count, translate[sspin])
    pin_menu = cw_cmenu.add_submenu(sitem)

    found1 = false
    all_aggregates.each { |e, s|
      next if s != 2
      if e.get_attribute(cw_gdict, 'Pin Pos O')
        found1 = true
        break
      end
    }

    if found1
      pin_menu.add_item(translate['Reset Position']) {
        op = 'CW - Reset Position'
        model.start_operation(op, true)
        all_aggregates.each { |e, s|
          next if s != 2
          orig_pos = e.get_attribute(cw_gdict, 'Pin Pos O')
          if orig_pos.is_a?(Array)
            e.set_attribute(cw_gdict, 'Pin Pos N', e.transformation.to_a)
            begin
              e.transformation = ::Geom::Transformation.new(orig_pos)
            rescue Exception => err
              puts err
              puts err.backtrace.first
            end
          end
          e.delete_attribute(cw_gdict, 'Pin Pos O')
        }
        model.commit_operation
      }
    end

    found2 = false
    all_aggregates.each { |e, s|
      next if s != 2
      if e.get_attribute(cw_gdict, 'Pin Pos N')
        found2 = true
        break
      end
    }

    if found2
      pin_menu.add_item(translate['Renew Position']) {
        op = 'CW - Renew Position'
        model.start_operation(op, true)
        all_aggregates.each { |e, s|
          next if s != 2
          new_pos = e.get_attribute(cw_gdict, 'Pin Pos N')
          if new_pos.is_a?(Array)
            e.set_attribute(cw_gdict, 'Pin Pos O', e.transformation.to_a)
            begin
              e.transformation = ::Geom::Transformation.new(new_pos)
            rescue Exception => err
              puts err
              puts err.backtrace.first
            end
          end
          e.delete_attribute(cw_gdict, 'Pin Pos N')
        }
        model.commit_operation
      }
    end

    if found1 || found2
      pin_menu.add_item(translate['Clear Memory']) {
        op = 'CW - Clear Pin History'
        model.start_operation(op, true)
        all_aggregates.each { |e, s|
          next if s != 2
          e.delete_attribute(cw_gdict, 'Pin Pos O')
          e.delete_attribute(cw_gdict, 'Pin Pos N')
        }
        model.commit_operation
      }
      pin_menu.add_separator
    end

    pin_menu.add_item(translate['Edit Scale']) {
      scale = AMS::ClothWorks.get_pin_scale
      prompts = [translate['Scale']]
      defaults = [sprintf("%0.2f", scale)]
      list = ['0.05|0.10|0.25|0.50|0.75|1.00|1.25|1.50|1.75|2.00|3.00|4.00|5.00|10.00']
      input = ::UI.inputbox(prompts, defaults, list, translate['Scale ClothWorks Pins'])
      next unless input
      scale = input[0].to_f
      op = 'CW - Edit Pins Scale'
      model.start_operation(op, true)
      begin
        model.set_attribute(cw_gdict, 'Pin Scale', scale)
        all_aggregates.each { |e, s|
          next if s != 2
          st = ::Geom::Transformation.scaling(scale)
          tra = e.transformation
          t = ::Geom::Transformation.new(tra.xaxis, tra.yaxis, tra.zaxis, tra.origin)
          e.transformation = t * st
        }
      rescue Exception => err
        model.abort_operation
        raise err
      end
      model.commit_operation
    }

    pin_menu.add_item(translate['Ignore']) {
      op = 'CW - Ignore'
      model.start_operation(op, true)
      all_aggregates.each { |e, s|
        next if s != 2
        e.delete_attribute(cw_gdict, 'Type')
      }
      model.commit_operation
      AMS::ClothWorks::UI.update_state(true)
    }
  end

  if voids_count > 0
    unless separator_added
      cw_cmenu.add_separator
      separator_added = true
    end

    ssvoid = voids_count > 1 ? "Voids" : "Void"
    sitem = sprintf("%d %s", voids_count, translate[ssvoid])
    void_menu = cw_cmenu.add_submenu(sitem)

    void_menu.add_item(translate['Ignore']) {
      op = 'CW - Ignore'
      model.start_operation(op, true)
      all_aggregates.each { |e, s|
        next if s != 0
        e.delete_attribute(cw_gdict, 'Type')
      }
      model.commit_operation
      AMS::ClothWorks::UI.update_state(true)
    }
  end
}

# Show toolbar
toolbar.show

end