# ----------------------------------------------------------------------------------------------------------------------
#
# Copyright (C) 2019 Anton Synytsia
#
# ----------------------------------------------------------------------------------------------------------------------

class AMS::ClothWorks::Simulation

  GIZMO_MIN_DIST_RATIO = 0.07

  SEL_BOX_COLOR1 = ::Sketchup::Color.new(100, 100, 100)
  SEL_BOX_COLOR2 = ::Sketchup::Color.new(255, 255, 100, 10)

  PIN_BOUNDS_COLOR = ::Sketchup::Color.new(214, 0, 211)
  PIN_BOUNDS_LINE_WIDTH = 1
  PIN_BOUNDS_LINE_STIPPLE = ''
  PIN_BOUNDS_OFFSET = 0.15

  PICKED_DRAW = {
    :line_width     => 1,
    :line_stipple   => '_',
    :line_color     => ::Sketchup::Color.new(250, 10, 10),
    :point_width    => 2,
    :point_size     => 10,
    :point_style    => 4,
    :point_color    => ::Sketchup::Color.new(4, 4, 4),
    :vline_width    => 1,
    :vline_stipple1 => '',
    :vline_stipple2 => '-',
    :vline_color    => ::Sketchup::Color.new(0, 40, 255)
  }

  SEL_PINS_DRAW = {
    :line_width     => 2,
    :line_stipple1  => '',
    :line_stipple2  => '_',
    :xline_color    => ::Sketchup::Color.new(235, 0, 0),
    :yline_color    => ::Sketchup::Color.new(0, 200, 0),
    :zline_color    => ::Sketchup::Color.new(0, 0, 200),
    :xline_colorh   => ::Sketchup::Color.new(255, 100, 100),
    :yline_colorh   => ::Sketchup::Color.new(20, 235, 20),
    :zline_colorh   => ::Sketchup::Color.new(0, 0, 255),
    :xline_colora   => ::Sketchup::Color.new(255, 0, 0),
    :yline_colora   => ::Sketchup::Color.new(0, 255, 0),
    :zline_colora   => ::Sketchup::Color.new(0, 255, 255),
    :bounds_width   => 2,
    :bounds_stipple => '_',
    :bounds_color   => ::Sketchup::Color.new(255, 127, 0),
    :padding        => ::Geom::Vector3d.new(0.1, 0.1, 0.1)
  }

  CLOTH_MESH_DRAW_OPTS = {
    :tcolor         => ::Sketchup::Color.new(255, 255, 255),
    :ecolor1        => ::Sketchup::Color.new(0, 0, 0),
    :ecolor2        => ::Sketchup::Color.new(0, 0, 0),
    :estipple1      => '-',
    :estipple2      => '',
    :ewidth1        => 1,
    :ewidth2        => 2,
  }

  SELECTED_CLOTH_MESH_DRAW_OPTS = {
    :tcolor         => ::Sketchup::Color.new(255, 255, 255),
    :ecolor1        => ::Sketchup::Color.new(0, 0, 255),
    :ecolor2        => ::Sketchup::Color.new(0, 0, 255),
    :estipple1      => '-',
    :estipple2      => '',
    :ewidth1        => 1,
    :ewidth2        => 2,
  }

  GIZMO_OPTS = {
    :line_length    => 0.9,
    :arrow_size     => 0.35,
    :scale_box_size => 0.15,
    :arc_radius     => 0.7,
    :point_radius   => 0.07,

    :txcolor        => ::Sketchup::Color.new(235, 0, 0),
    :tycolor        => ::Sketchup::Color.new(0, 200, 0),
    :tzcolor        => ::Sketchup::Color.new(0, 0, 200),
    :txcolorh       => ::Sketchup::Color.new(255, 100, 100),
    :tycolorh       => ::Sketchup::Color.new(20, 235, 20),
    :tzcolorh       => ::Sketchup::Color.new(0, 0, 255),
    :txcolora       => ::Sketchup::Color.new(255, 100, 100),
    :tycolora       => ::Sketchup::Color.new(20, 235, 20),
    :tzcolora       => ::Sketchup::Color.new(0, 0, 255),
    :twidth         => 2,
    :tstipple       => '',
    :twidthh        => 2,
    :tstippleh      => '',
    :twidtha        => 2,
    :tstipplea      => '',

    :sxcolor        => ::Sketchup::Color.new(235, 0, 0),
    :sycolor        => ::Sketchup::Color.new(0, 200, 0),
    :szcolor        => ::Sketchup::Color.new(0, 0, 200),
    :sxcolorh       => ::Sketchup::Color.new(255, 100, 100),
    :sycolorh       => ::Sketchup::Color.new(20, 235, 20),
    :szcolorh       => ::Sketchup::Color.new(0, 0, 255),
    :sxcolora       => ::Sketchup::Color.new(255, 100, 100),
    :sycolora       => ::Sketchup::Color.new(20, 235, 20),
    :szcolora       => ::Sketchup::Color.new(0, 0, 255),
    :swidth         => 1,
    :sstipple       => '_',
    :swidthh        => 1,
    :sstippleh      => '_',
    :swidtha        => 1,
    :sstipplea      => '_',

    :rcolor         => ::Sketchup::Color.new(255, 255, 0),
    :rcolorh        => ::Sketchup::Color.new(255, 127, 0),
    :rcolora        => ::Sketchup::Color.new(255, 255, 0),
    :rwidth         => 1,
    :rstipple       => '',
    :rwidthh        => 1,
    :rstippleh      => '',
    :rwidtha        => 1,
    :rstipplea      => '',
    :rnum_seg       => 64,

    :pcolorh        => ::Sketchup::Color.new(255, 255, 0),
    :pcolora        => ::Sketchup::Color.new(255, 127, 0),
    :pnum_seg       => 32
  }

  BOX_SPACE_OPTS = {
    :color          => ::Sketchup::Color.new(255, 127, 0),
    :width          => 1,
    :stipple        => '',
  }

  POINT_OPTS = {
    :color          => ::Sketchup::Color.new(0, 255, 255),
    :width          => 1,
    :stipple        => '',
    :size           => 4,
    :style          => 2
  }

  NORMAL_OPTS = {
    :color          => ::Sketchup::Color.new(255, 0, 0),
    :width          => 1,
    :stipple        => '',
    :scale          => 3
  }

  def initialize(p_model, bss = false)
    unless p_model.is_a?(::Sketchup::Model)
      raise(TypeError, 'Expected a Sketchup::Model instance!')
    end

    @model = p_model
    @world = nil
    @frame = 0
    @animation_stopped = false
    @colliders = {}
    @cloths = {}
    @pins = {}
    @voids = {}
    @pick_mouse_moved = false
    @pick_plane_origin = nil
    @sel_ents = {}
    @sel_pins = {}
    @first_sel_pin = nil
    @sel_pins_dat = {}
    @sel_box_pt = nil
    @sel_box_active = false
    @sel_ent = nil
    @paused = false
    @pause_updated = false
    @suspended = false
    @mouse_over = false
    @fps = {
      :elapsed  => 0,
      :last     => ::Time.now,
      :rate     => 20,
      :result   => 0
    }
    @ip = ::Sketchup::InputPoint.new
    @cursor_pos = ::Geom::Point3d.new(0,0,0)
    @bb = ::Geom::BoundingBox.new

    @key_state = {
      :left     => 0,
      :right    => 0,
      :up       => 0,
      :down     => 0,
      :shift    => 0,
      :control  => 0,
      :home     => 0,
      :end      => 0,
      :cdp      => 0
    }

    @scene_info = {
      :active => false,
      :data1 => nil,
      :data2 => nil,
      :transition_time => 0,
      :elasted_time => 0,
      :timer => nil
    }

    @scene_anim_info = {
      :state => 0,
      :data => {},
      :transition_time => 0,
      :elasted_time => 0,
      :ref_time => 0,
      :tabs_size => nil,
      :active_tab => nil,
      :tab_dir => 1
    }

    @frame_change_observer_id = nil
    @ams_su_observer_added = false

    valid_deformables = {}

    # Fix meshes
    op = 'CW - Fix Meshes'
    @model.start_operation(op, true)

    cw_gdict = AMS::ClothWorks::CW_GDICT

    begin
      # Close active path
      if @model.active_entities != @model.entities
        state = true
        while state
          state = @model.close_active
        end
      end

      # Fix deformable meshes prior to starting simulation
      instances_to_fix = []
      @model.entities.to_a.each { |entity|
        next if !entity.is_a?(::Sketchup::Group) && !entity.is_a?(::Sketchup::ComponentInstance)
        next if !entity.visible? || !entity.layer.visible?
        next if bss && !@model.selection.include?(entity)
        type = entity.get_attribute(cw_gdict, 'Type', nil)
        next if !type.is_a?(Integer)
        if type == 0
          definition = AMS::ClothWorks.get_definition(entity)
          behavior = definition.behavior
          behavior.always_face_camera = false if behavior.always_face_camera?
        elsif type == 1
          definition = AMS::ClothWorks.get_definition(entity)
          entities = definition.entities
          if AMS::ClothWorks.faces_triangular?(entities)
            behavior = definition.behavior
            behavior.always_face_camera = false if behavior.always_face_camera?
            entity.make_unique # This is important
            valid_deformables[entity] = true
          else
            instances_to_fix << entity
          end
        end
      }

      created_instances = AMS::ClothWorks.triangulate_faces_multiple(instances_to_fix)
      created_instances.each { |oe, ne|
        oe.erase! if oe.valid?
        ne.make_unique # This is important
        valid_deformables[ne] = true
      }

      @model.commit_operation
    rescue Exception => err
      @model.abort_operation
      raise err
    end

    # Create world
    @world = AMS::ClothWorks::World.new(AMS::ClothWorks.solver)

    lr = AMS::ClothWorks.licensed?
    ct = AMS::ClothWorks.time
    @ulf = true

    # Add cloths, colliders, pins, and voids
    cmrs = {} # { definition => mesh } - used for cloning as opposed to re-reading identical collider meshes

    @model.entities.to_a.each { |entity|
      next if !entity.is_a?(::Sketchup::Group) && !entity.is_a?(::Sketchup::ComponentInstance)

      type = entity.get_attribute(cw_gdict, 'Type', nil)
      next unless type.is_a?(Integer)

      if entity.is_a?(::Sketchup::Group)
        entities = entity.entities
        definition = AMS::ClothWorks.get_definition(entity)
      else
        definition = entity.definition
        entities = definition.entities
      end
      transformation = entity.transformation

      if type == 0 # Collider
        next if !entity.visible? || !entity.layer.visible?

        def_tras = AMS::ClothWorks.get_definitions_transformations(entity, transformation) { |e|
          e.layer.visible? && e.visible? && !e.get_attribute(cw_gdict, 'Exclude')
        }
        next if def_tras.empty?

        collider = AMS::ClothWorks::Collider.new
        @world.add_collider(collider)

        def_tras.each { |d, tras|
          tras.each { |t|
            cmr = cmrs[d]
            if cmr
              collider_mesh = cmr.dup
              collider_mesh.transformation = t
            else
              collider_mesh = AMS::ClothWorks::ColliderMesh.new(d, t)
              cmrs[d] = collider_mesh
            end
            collider.add_mesh(collider_mesh)
          }
        }

        val = entity.get_attribute(cw_gdict, 'Thickness', AMS::ClothWorks::DEFAULT_COLLIDER_THICKNESS).to_f
        collider.thickness = val * AMS::ClothWorks::METER_TO_INCH
        val = entity.get_attribute(cw_gdict, 'Friction', AMS::ClothWorks::DEFAULT_COLLIDER_FRICTION).to_f
        collider.friction = val

        @colliders[entity] = collider

      elsif type == 1 # Cloth
        next if !entity.visible? || !entity.layer.visible? || !valid_deformables.has_key?(entity)

        cloth = AMS::ClothWorks::Cloth.new
        world.add_cloth(cloth)

        # Create
        drape_state = definition.get_attribute(cw_gdict, 'State')
        cloth_mesh = AMS::ClothWorks::ClothMesh.new(entity, transformation, drape_state, true)
        cloth.add_mesh(cloth_mesh)

        # Set quality
        quality = entity.get_attribute(cw_gdict, 'Quality', AMS::ClothWorks::DEFAULT_QUALITY).to_i
        cloth.quality = quality
        cloth.min_refine_angle = AMS::ClothWorks::MIN_REFINE_ANGLE
        cloth.max_refine_angle = AMS::ClothWorks::MAX_REFINE_ANGLE
        cloth.coarse_ratio = AMS::ClothWorks::COARSE_RATIO

        # Add cross edges
        gen_cross = entity.get_attribute(cw_gdict, 'Generate Shear', AMS::ClothWorks::DEFAULT_GENERATE_SHEAR)
        if gen_cross && quality < 0
          cloth_mesh.generate_cross_edges
        end

        # Add edge links
        cloth_mesh.generate_edge_links

        # Add triplet links
        cloth_mesh.generate_triplet_links

        # Update properties
        mat = AMS::ClothWorks.get_cloth_material(entity)
        cloth.density = mat.density
        cloth.thickness = mat.thickness * AMS::ClothWorks::METER_TO_INCH
        cloth.friction = mat.friction
        cloth.stretch_stiffness = mat.stretch_stiffness
        cloth.bend_stiffness = mat.bend_stiffness
        cloth.damper = mat.damper
        cloth.scale = mat.scale
        cloth.lonely_scale = mat.lonely_scale
        cloth.drag = mat.drag?
        cloth.self_collide = mat.self_collide?
        cloth.preserve_bendings = mat.preserve_bendings?
        cloth.update_scale_factors

        # Store
        @cloths[entity] = cloth

      elsif type == 2 # Pin
        box_datas = AMS::ClothWorks.get_aligned_boxes(entity, true, nil) { |e|
          e.visible?
        }
        begin
          orig_tra = entity.get_attribute(cw_gdict, 'Pin Pos O')
          if orig_tra.is_a?(Array)
            init_transformation = ::Geom::Transformation.new(orig_tra)
          else
            init_transformation = transformation
          end

          pin = AMS::ClothWorks::Pin.new(init_transformation, transformation, entity)
          @world.add_pin(pin)

          box_datas.each { |box_data|
            pin.add_box(box_data[1], box_data[2])
          }
        rescue TypeError => err
          puts err
          puts err.backtrace.first
          next
        end

        # Update properties
        val = entity.get_attribute(cw_gdict, 'Padding', AMS::ClothWorks::DEFAULT_PIN_PADDING).to_f
        pin.padding = val * AMS::ClothWorks::METER_TO_INCH

        # Store
        @pins[entity] = pin

      elsif type == 3 # Void
        box_datas = AMS::ClothWorks.get_aligned_boxes(entity, true, nil) { |e|
          e.visible?
        }

        begin
          vb = AMS::ClothWorks::Void.new(transformation)
          @world.add_void(vb)

          box_datas.each { |box_data|
            vb.add_box(box_data[1], box_data[2])
          }
        rescue TypeError => err
          puts err
          puts err.backtrace.first
          next
        end

        # Store
        @voids[entity] = vb

      end
    }

    # Finalize world
    @world.finalize(AMS::ClothWorks::PATH)

    # Pin down vertices
    #@world.pin_initially_overlapping_vertices
    @world.pin_overlapping_vertices

    # Update mass factors
    @world.update_mass_factors

    # Update world properties
    @world.gravity = ::Geom::Vector3d.new(0, 0, -AMS::ClothWorks.gravity)
    @world.timestep = AMS::ClothWorks.update_timestep
    @world.drag = AMS::ClothWorks.drag
    @world.viscosity = AMS::ClothWorks.viscosity
    @world.wind_velocity = ::Geom::Vector3d.new(
      AMS::ClothWorks.wind_velocity_x,
      AMS::ClothWorks.wind_velocity_y,
      AMS::ClothWorks.wind_velocity_z)
    @world.grab_stiffness = AMS::ClothWorks::PICK_ACCEL
    @world.grab_damper = AMS::ClothWorks::PICK_DAMP

  end

  attr_reader :frame, :world, :model, :cloths, :colliders, :pins

  def update_status_text
    if @mouse_over && !@suspended
      tr = AMS::ClothWorks.translate
      txt = sprintf("%s: %d   FPS: %d   %s: %.2f s   %s   %s",
        tr["Frame"], @frame,
        @fps[:result],
        tr["Time"], @world.elapsed_time,
        tr["Click and drag to move cloth."],
        tr["Use ARROW, HOME, END keys to move selected pins; hold SHIFT key to converge/diverge."])
      ::Sketchup.set_status_text(txt, SB_PROMPT)
    end
  end

  def update_selection
    sel = @model.selection
    sel.clear
    consider_hidden = @model.rendering_options['DrawHidden']
    @sel_pins.clear
    @sel_pins_dat.clear
    @sel_ents.reject! { |e, s|
      next true unless e.valid?
      if e.layer.visible? && (e.visible? || consider_hidden || (AMS::ClothWorks.draw_mesh? && @cloths.has_key?(e)))
        sel.add(e)
        pin = @pins[e]
        next false unless pin
        if @ulf
          @sel_pins[pin] = []
        end
        false
      else
        true
      end
    }
    if @sel_pins.empty?
      @first_sel_pin = nil
    elsif @first_sel_pin.nil? || (@first_sel_pin && !@sel_pins.has_key?(@first_sel_pin))
      @first_sel_pin = @sel_pins.keys.first
    end
    self.update_sel_pins_bounds
    AMS::ClothWorks::UI.update_state(true)
  end

  def complete(reset_positions = false)
    return unless @world

    # Reset cloth and pins
    op = 'CW - End Simulation'
    @model.start_operation(op, true)

    cw_gdict = AMS::ClothWorks::CW_GDICT

    begin
      if reset_positions
        @world.cloths.each { |cloth|
          cloth.meshes.each { |mesh|
            next unless mesh.su_definition.valid?
            mesh.transform_vertices_to_initial_session
            mesh.position_initial_session_materials

            if mesh.su_instance.valid? && mesh.su_instance.hidden?
              mesh.su_instance.visible = true
            end
          }
        }

        @world.pins.each { |pin|
          next unless pin.su_instance.valid?
          pin.su_instance.move!(pin.initial_session_transformation)
        }

      else
        @world.cloths.each { |cloth|
          cloth.meshes.each { |mesh|
            next unless mesh.su_definition.valid?
            mesh.transform_vertices
            mesh.position_materials
            loc_init_presets = mesh.local_initial_vertex_positions
            loc_orig_presets = mesh.local_original_vertex_positions
            AMS::ClothWorks.store_draped_vertex_positions(mesh.su_definition, loc_init_presets, loc_orig_presets)

            if mesh.su_instance.valid? && mesh.su_instance.hidden?
              mesh.su_instance.visible = true
            end
          }
        }

        @world.pins.each { |pin|
          e = pin.su_instance
          next unless e.valid?
          cur_data = e.get_attribute(cw_gdict, 'Pin Pos O')
          e.delete_attribute(cw_gdict, 'Pin Pos N')
          next if cur_data.is_a?(Array)
          orig_tra = pin.initial_session_transformation
          if e.transformation.origin.distance(orig_tra.origin).to_f > AMS::ClothWorks::EPSILON
            e.set_attribute(cw_gdict, 'Pin Pos O', orig_tra.to_a)
          end
        }
      end

      @model.commit_operation
    rescue Exception => err
      @model.abort_operation
      raise err
    end

    # Reset
    @world.reset

    # Generate subdivided meshes
    if !reset_positions
      op = 'CW - Spawn Subdivided Meshes'
      @model.start_operation(op, true)

      cw_gdict = AMS::ClothWorks::CW_GDICT

      begin
        @world.cloths.each { |cloth|
          next if cloth.quality < 0

          cloth.meshes.each { |mesh|
            # Spawn new
            g = @model.entities.add_group
            g.move!(mesh.transformation)
            vc = mesh.fill_entities(g.entities, mesh.transformation.inverse, true, 1, true)
            if vc > 0
              d = AMS::ClothWorks.get_definition(g)
              d.set_attribute(cw_gdict, 'State', 1)
              d.set_attribute(cw_gdict, 'Count', vc)
            end

            # Reassign attributes
            if mesh.su_instance.valid?
              re = mesh.su_instance
              g.name = re.name
              g.layer = re.layer
              g.material = re.material
              g.visible = re.visible?
              dicts = re.attribute_dictionaries
              if dicts
                dicts.each { |dict|
                  dname = dict.name
                  dict.each { |k, v|
                    g.set_attribute(dname, k, v)
                  }
                }
              end
              #g.set_attribute(cw_gdict, 'Quality', AMS::ClothWorks::DEFAULT_QUALITY)
            end

            # Destroy original mesh
            if mesh.su_instance.valid?
              mesh.su_instance.erase!
            end
          }
        }

        @model.commit_operation
      rescue Exception => err
        @model.abort_operation
        raise err
      end
    end

    # Purge
    op = 'CW - Purge Definitions'
    @model.start_operation(op, true)

    @model.definitions.purge_unused
    @model.commit_operation

    # Clear variables
    @colliders.clear
    @cloths.clear
    @pins.clear
    @voids.clear
    @sel_ents.clear
    @sel_pins.clear
    @sel_pins_dat.clear
    @scene_anim_info.clear
    @scene_info.clear
    @sel_box_pt = nil
    @sel_box_active = false
    @sel_ent = nil
    @world = nil

    # Start garbage collection
    ::GC.start

  end

  def sync_cloth_properties(inst)
    cloth = @cloths[inst]
    return unless cloth
    mat = AMS::ClothWorks.get_cloth_material(inst)
    cloth.density = mat.density
    cloth.thickness = mat.thickness * AMS::ClothWorks::METER_TO_INCH
    cloth.friction = mat.friction
    cloth.stretch_stiffness = mat.stretch_stiffness
    cloth.bend_stiffness = mat.bend_stiffness
    cloth.damper = mat.damper
    cloth.scale = mat.scale
    cloth.lonely_scale = mat.lonely_scale
    cloth.drag = mat.drag?
    cloth.self_collide = mat.self_collide?
    cloth.preserve_bendings = mat.preserve_bendings?
    cloth.update_scale_factors
    cloth.update_mass_factors
  end

  def advance_to_next_frame(view)
    return unless @world

    # Update world
    @world.advance(AMS::ClothWorks.update_rate)

    # Update cloths
    if AMS::ClothWorks.draw_mesh?
      @world.cloths.each { |cloth|
        cloth.meshes.each { |mesh|
          if mesh.su_instance.valid? && mesh.su_instance.visible?
            mesh.su_instance.visible = false
          end
        }
      }
    else
      @world.cloths.each { |cloth|
        cloth.meshes.each { |mesh|
          next unless mesh.su_definition.valid?
          mesh.transform_vertices
          if AMS::ClothWorks.position_textures?
            mesh.position_materials
          end

          if mesh.su_instance.valid? && mesh.su_instance.hidden?
            mesh.su_instance.visible = true
          end
        }
      }
    end

    # Update pins
    @world.pins.each { |pin|
      if pin.su_instance.valid?
        pin.su_instance.move!(pin.transformation)
      end
    }

    # Record
    if AMS::ClothWorks.record? && !AMS::ClothWorks.record_instance.nil?
      AMS::ClothWorks.record_instance.record_world(@world, @frame)
    end

    # Set console
    # unless AMS::ClothWorks::Console.open?
      # AMS::ClothWorks::Console.open
    # end
    # AMS::ClothWorks::Console.set(AMS::ClothWorks.profile)

    # Update frame
    @frame += 1
  end

  def paused?
    @paused
  end

  def paused=(state)
    @paused = state ? true : false
  end

  def activate
    if AMS::ClothWorks::SU_MAJOR_VERSION > 6
      ::Sketchup.break_edges = false
    end
    @model.active_view.animation = self
    # Activate page observer
    @frame_change_observer_id = ::Sketchup::Pages.add_frame_change_observer(self)
    self.animate_scenes(AMS::ClothWorks.animate_scenes_state)
    # Activate observer
    if AMS::ClothWorks::IS_PLATFORM_WINDOWS && defined?(AMS::Lib) && defined?(AMS::Lib::VERSION) && AMS::Lib::VERSION.to_f >= 3.5
      AMS::Sketchup.add_observer(self)
      @ams_su_observer_added = true
    end
  end

  def deactivate(view)
    # Remove AMS SU observer
    if @ams_su_observer_added
      AMS::Sketchup.remove_observer(self)
      @ams_su_observer_added = false
    end
    # Deactivate page observer (for OS X)
    if @frame_change_observer_id
      ::Sketchup::Pages.remove_frame_change_observer(@frame_change_observer_id)
      @frame_change_observer_id = nil
    end
    complete(false) if @world
    view.animation = nil
    if AMS::ClothWorks::SU_MAJOR_VERSION > 6
      ::Sketchup.break_edges = true
    end
    # Validate record
    if AMS::ClothWorks.record_instance && AMS::ClothWorks.record_instance.data_empty?
      AMS::ClothWorks.clear_record
    end
    # Refresh view
    view.invalidate
    AMS::ClothWorks::UI.update_state(false)
  end

  def onCancel(reason, view)
    @model.select_tool(nil)
  end

  def onMouseEnter(view)
    @mouse_over = true
    self.update_status_text
  end

  def onMouseLeave(view)
    @mouse_over = false
  end

  def onMouseMove(flags, x, y, view)
    @cursor_pos.x = x
    @cursor_pos.y = y
    @sel_box_active = true

    if @world.grab_active?
      @pick_mouse_moved = true
    end
  end

  def resume(view)
    @suspended = false
    if @world.grab_active?
      @pick_plane_origin = @world.grab_destination_point
    end
    self.update_status_text
  end

  def suspend(view)
    @suspended = true
  end

  def onLButtonDown(flags, x, y, view)
    # Release grab
    @world.release_grab

    # Clear selection data
    @sel_box_pt = nil
    @sel_box_active = false
    @sel_ent = nil
    @sel_pins_dat[:sid] = nil
    @sel_pins_dat[:pick_pt] = nil

    # Select gizmo
    if !@paused && !@sel_pins.empty?
      c = @sel_pins_dat[:tra1].origin#@sel_pins_dat[:bb].center
      d = @sel_pins_dat[:size]
      cdist = c.distance(view.camera.eye).to_f
      if cdist > AMS::ClothWorks::EPSILON && d / cdist < GIZMO_MIN_DIST_RATIO
        d = cdist * GIZMO_MIN_DIST_RATIO
      end

      info = AMS::ClothWorks.select_gizmo(view, x, y, @sel_pins_dat[:tra1], d, GIZMO_OPTS)
      if info
        @sel_pins_dat[:sid] = info[0]
        @sel_pins_dat[:pick_pt] = info[1]
        @sel_pins_dat[:init_scale] = ::Geom::Vector3d.new(@sel_pins_dat[:scale])

        @sel_pins_dat[:vel] = ::Geom::Vector3d.new(0,0,0)
        @sel_pins_dat[:omega] = ::Geom::Vector3d.new(0,0,0)
        @sel_pins_dat[:scale_vel] = ::Geom::Vector3d.new(0,0,0)
        return
      end
    end

    # Artificial cloth selection
    ray = view.pickray(x,y)
    if AMS::ClothWorks.draw_mesh?
      cloth_mesh = @world.select_cloth_mesh(ray[0], ray[1])
      if cloth_mesh && cloth_mesh.su_instance.valid?
        @sel_ent = cloth_mesh.su_instance
      end
    end

    # Entity selection
    unless @sel_ent
      consider_hidden = view.model.rendering_options['DrawHidden']

      ph = view.pick_helper
      ph.do_pick(x,y)

      element = ph.best_picked
      sel_face = ph.picked_face
      sel_edge = ph.picked_edge

      if element && (AMS::ClothWorks::SU_MAJOR_VERSION < 16 || !element.is_a?(::Sketchup::Axes)) && (consider_hidden || element.visible?)
        @sel_ent = element
      elsif sel_face && (consider_hidden || sel_face.visible?)
        @sel_ent = sel_face
      elsif sel_edge && (consider_hidden || sel_edge.visible?)
        @sel_ent = sel_edge
      end
    end

    # Box selection
    @sel_box_pt = ::Geom::Point3d.new(x, y, 0)

    # Pick and drag
    @world.perform_grab(ray[0], ray[1])
    @pick_mouse_moved = false
    if @world.grab_active?
      @world.grab_destination_point = @world.grab_pick_point
      @pick_plane_origin = @world.grab_destination_point
    else
      @pick_plane_origin = nil
    end
  end

  def onLButtonDoubleClick(flags, x, y, view)
    onLButtonDown(flags, x, y, view)
  end

  def onLButtonUp(flags, x, y, view)
    # If selected gizmo, deselect gizmo and return
    if @sel_pins_dat[:sid]
      @sel_pins_dat[:tra1] = @sel_pins_dat[:tra2]
      @sel_pins_dat[:sid] = nil
      @sel_pins_dat[:pick_pt] = nil
      return
    end

    # Process seleciton
    if !@world.grab_active? || !@pick_mouse_moved
      if @sel_box_pt && @sel_box_active # Box selection
        if AMS::ClothWorks::SU_MAJOR_VERSION > 15
          ph = view.pick_helper
          ph.window_pick(@sel_box_pt, ::Geom::Point3d.new(x,y,0), @sel_box_pt.x < x ? ::Sketchup::PickHelper::PICK_INSIDE : ::Sketchup::PickHelper::PICK_CROSSING)
          ents = ph.all_picked
        else
          consider_hidden = view.model.rendering_options['DrawHidden']
          if @sel_box_pt.x < x
            ents = AMS::ClothWorks.get_entities_within_selection_box(view, @sel_box_pt, ::Geom::Point3d.new(x,y,0), consider_hidden)
          else
            ents = AMS::ClothWorks.get_entities_overlapping_selection_box(view, @sel_box_pt, ::Geom::Point3d.new(x,y,0), consider_hidden)
          end
        end

        if @key_state[:control] == 1 && @key_state[:shift] == 1
          ents.each { |e|
            @sel_ents.delete(e)
          }
        elsif @key_state[:control] == 1
          ents.each { |e|
            @sel_ents[e] = true
          }
        elsif @key_state[:shift] == 1
          ents.each { |e|
            if @sel_ents[e]
              @sel_ents.delete(e)
            else
              @sel_ents[e] = true
            end
          }
        else
          @sel_ents.clear
          ents.each { |e|
            @sel_ents[e] = true
          }
        end
      elsif @sel_ent && @sel_ent.valid? # Individual entity selection
        if @key_state[:control] == 1 && @key_state[:shift] == 1
          @sel_ents.delete(@sel_ent)
        elsif @key_state[:control] == 1
          @sel_ents[@sel_ent] = true
        elsif @key_state[:shift] == 1
          if @sel_ents[@sel_ent]
            @sel_ents.delete(@sel_ent)
          else
            @sel_ents[@sel_ent] = true
          end
        else
          @sel_ents.clear
          @sel_ents[@sel_ent] = true
        end
      else # Empty space selected
        @sel_ents.clear
      end
      self.update_selection
    end

    # Release grab
    @world.release_grab

    # Clear selection data
    @sel_box_pt = nil
    @sel_box_active = false
    @sel_ent = nil
  end

  def getMenu(menu)
    view = @model.active_view
    ph = view.pick_helper
    ph.do_pick(@cursor_pos.x, @cursor_pos.y)
    ent = ph.best_picked
    if ent.nil? || (AMS::ClothWorks::SU_MAJOR_VERSION > 15 && ent.is_a?(::Sketchup::Axes))
      @sel_ents.clear
      self.update_selection
      return
    end
    unless @sel_ents[ent]
      @sel_ents.clear
      @sel_ents[ent] = true
      self.update_selection
    end
    menu.add_item(@paused ? 'Play' : 'Pause') {
      @paused = !@paused
    }
    menu.add_item('Stop') {
      complete(false) if @world
      @model.select_tool nil
    }
    menu.add_item('Reset') {
      complete(true) if @world
      @model.select_tool nil
    }

    menu.add_separator

    bshow = false
    @sel_ents.each { |e, s|
      next unless e.valid?
      unless e.visible?
        bshow = true
        break
      end
    }

    if bshow
      menu.add_item(AMS::ClothWorks.translate['Show']) {
        @sel_ents.each { |e, s|
          next unless e.valid?
          e.visible = true unless e.visible?
        }
        self.update_selection
      }

    else
      menu.add_item(AMS::ClothWorks.translate['Hide']) {
        @sel_ents.each { |e, s|
          next unless e.valid?
          e.visible = false if e.visible?
        }
        self.update_selection
      }
    end
  end

  def onKeyDown(key, repeat, flags, view)
    case key
      when VK_LEFT
        @key_state[:left] = 1
      when VK_RIGHT
        @key_state[:right] = 1
      when VK_UP
        @key_state[:up] = 1
      when VK_DOWN
        @key_state[:down] = 1
      when VK_SHIFT
        @key_state[:shift] = 1
      when VK_CONTROL
        @key_state[:control] = 1
      when VK_HOME
        @key_state[:home] = 1
      when VK_END
        @key_state[:end] = 1
    end
    @key_state[:cdp] = 0
    if AMS::ClothWorks::IS_PLATFORM_WINDOWS
      if @key_state[:shift] == 1 # scale
        @key_state[:cdp] = 1
      elsif @key_state[:control] == 1 # rotate
        @key_state[:cdp] = 2
      end
    elsif @key_state[:shift] == 1
      @key_state[:cdp] = 1
    end
  end

  def onKeyUp(key, repeat, flags, view)
    case key
      when VK_LEFT
        @key_state[:left] = 0
      when VK_RIGHT
        @key_state[:right] = 0
      when VK_UP
        @key_state[:up] = 0
      when VK_DOWN
        @key_state[:down] = 0
      when VK_SHIFT
        @key_state[:shift] = 0
      when VK_CONTROL
        @key_state[:control] = 0
      when VK_HOME
        @key_state[:home] = 0
      when VK_END
        @key_state[:end] = 0
    end
    @key_state[:cdp] = 0
    if AMS::ClothWorks::IS_PLATFORM_WINDOWS
      if @key_state[:shift] == 1 # scale
        @key_state[:cdp] = 1
      elsif @key_state[:control] == 1 # rotate
        @key_state[:cdp] = 2
      end
    elsif @key_state[:shift] == 1
      @key_state[:cdp] = 1
    end
  end

  def getInstructorContentDirectory
  end

  def getExtents
    bb = ::Geom::BoundingBox.new()
    bb.add(@bb)
    @bb.clear
    bb
  end

  def nextFrame(view)
    return false unless @world

    if @paused
      unless @pause_updated
        @fps[:elapsed] += Time.now - @fps[:last]
        @pause_updated = true
      end
      self.update_status_text()
      view.show_frame(0.1)

    else
      if @pause_updated
        @fps[:last] = Time.now
        @pause_updated = false
      end

      # Wrap operation
      op = 'CW - Frame ' + frame.to_s
      @model.start_operation(op, true)

      begin
        # Process pick and drag
        if @world.grab_active?
          ray = view.pickray(@cursor_pos.x, @cursor_pos.y)
          plane = [@pick_plane_origin, view.camera.zaxis]
          @world.grab_destination_point = ::Geom.intersect_line_plane(ray, plane)
        end

        # Process selected pins
        self.update_sel_pins(view)

        # Advance simulation
        self.advance_to_next_frame(view)

        # Update selected pins bounds information
        self.update_sel_pins_bounds()

        # Update FPS
        if @frame %  @fps[:rate] == 0
          @fps[:elapsed] += Time.now - @fps[:last]
          if @fps[:elapsed] > 1.0e-6
            elapsed = @fps[:elapsed].to_f
            @fps[:result] = (@fps[:rate] / elapsed).round
          else
            @fps[:result] = 0
          end
          @fps[:last] = Time.now
          @fps[:elapsed] = 0
        end

        # Update camera
        self.update_scenes_animation() || self.update_scene_transitioning()

        # End operation
        @model.commit_operation
      rescue Exception => err
        # Abort operation
        @model.abort_operation
        raise err
      end

      # Update status
      self.update_status_text

      # Show frame
      view.show_frame

    end

    return true
  end

  def stop
    return unless @world
    @animation_stopped = true
    @model.active_view.invalidate
  end

  def draw_pick_and_drag(view)
    return unless @world.grab_active?

    pt1 = @world.grab_pick_point
    pt2 = @world.grab_destination_point

    @bb.add(pt1)
    @bb.add(pt2)

    view.line_width = PICKED_DRAW[:line_width]
    view.line_stipple = PICKED_DRAW[:line_stipple]
    view.drawing_color = PICKED_DRAW[:line_color]
    view.draw_line(pt1, pt2)

    view.line_width = PICKED_DRAW[:point_width]
    view.line_stipple = ''
    view.draw_points(pt1, PICKED_DRAW[:point_size], PICKED_DRAW[:point_style], PICKED_DRAW[:point_color])

    # if @key_state[:shift] == 1
      # view.line_width = PICKED_DRAW[:vline_width]
      # view.line_stipple = pt2.z > 0 ? PICKED_DRAW[:vline_stipple1] : PICKED_DRAW[:vline_stipple2]
      # view.drawing_color = PICKED_DRAW[:vline_color]
      # view.draw(GL_LINES, [pt2.x, pt2.y, 0], pt2)
    # end
  end

  def draw_cloth_meshes(view)
    if AMS::ClothWorks.draw_mesh?
      @world.cloths.each { |cloth|
        cloth.meshes.each { |mesh|
          if @sel_ents.has_key?(mesh.su_instance)
            bb = mesh.draw(view, SELECTED_CLOTH_MESH_DRAW_OPTS)
          else
            bb = mesh.draw(view, CLOTH_MESH_DRAW_OPTS)
          end
          # mesh.draw(view, CLOTH_MESH_DRAW_OPTS)
          @bb.add(bb)
        }
      }
    end
    if AMS::ClothWorks.draw_points?
      @world.cloths.each { |cloth|
        cloth.meshes.each { |mesh|
          bb = mesh.draw_points(view, POINT_OPTS[:color], POINT_OPTS[:size], POINT_OPTS[:style], POINT_OPTS[:width], POINT_OPTS[:stipple])
          @bb.add(bb)
        }
      }
    end
    if AMS::ClothWorks.draw_normals?
      @world.cloths.each { |cloth|
        cloth.meshes.each { |mesh|
          #bb = mesh.draw_normals(view, NORMAL_OPTS[:scale], NORMAL_OPTS[:color], NORMAL_OPTS[:width], NORMAL_OPTS[:stipple])
          bb = mesh.draw_vertex_normals(view, NORMAL_OPTS[:scale], NORMAL_OPTS[:color], NORMAL_OPTS[:width], NORMAL_OPTS[:stipple])
          @bb.add(bb)
        }
      }
    end
    if AMS::ClothWorks.draw_box_space?
      @world.cloths.each { |cloth|
        cloth.meshes.each { |mesh|
          bb = mesh.draw_vspace(view, BOX_SPACE_OPTS[:color], BOX_SPACE_OPTS[:width], BOX_SPACE_OPTS[:stipple])
          @bb.add(bb)
        }
      }
    end
  end

  def update_sel_pins(view)
    return if @sel_pins.empty?

    # Validate
    @sel_pins.reject! { |pin, data|
      next !pin.su_instance.valid?
    }

    if @sel_pins.empty?
      @sel_pins_dat.clear
      return
    end

    if @sel_pins_dat[:sid] && @sel_pins_dat[:pick_pt] # Mouse control
      info = AMS::ClothWorks.transform_gizmo(view, @cursor_pos.x, @cursor_pos.y, @sel_pins_dat[:tra1], @sel_pins_dat[:sid], @sel_pins_dat[:pick_pt])
      if info
        if @sel_pins_dat[:sid] >= 4 && @sel_pins_dat[:sid] <= 6 # converge/divege
          ds = @sel_pins_dat[:init_scale].dot(info[0]) * info[1]
          cs = @sel_pins_dat[:scale].dot(info[0])
          sf = AMS::ClothWorks.scale_vector(info[0], ds - cs)
          @sel_pins_dat[:scale] += sf
        else # rotation and translation
          @sel_pins_dat[:tra2] = info
        end
      end

    else # Keyboard control
      lin_acc = ::Geom::Vector3d.new(0,0,0)
      ang_acc = ::Geom::Vector3d.new(0,0,0)
      sca_acc = ::Geom::Vector3d.new(0,0,0)
      kx = @key_state[:right] - @key_state[:left]
      ky = @key_state[:up] - @key_state[:down]
      kz = @key_state[:home] - @key_state[:end]
      dr = ::Geom::Vector3d.new(kx, ky, kz)
      af = AMS::ClothWorks.pin_accel
      df = AMS::ClothWorks.pin_damp
      dt = AMS::ClothWorks::VIEW_UPDATE_TIMESTEP

      if @key_state[:cdp] == 1 # converge/diverge
        sca_acc = AMS::ClothWorks.scale_vector(dr, af)
        if kx != 0 && ky == 0 && kz == 0
          @sel_pins_dat[:sid] = 4
        elsif kx == 0 && ky != 0 && kz == 0
          @sel_pins_dat[:sid] = 5
        elsif kx == 0 && ky == 0 && kz != 0
          @sel_pins_dat[:sid] = 6
        else
          @sel_pins_dat[:sid] = nil
        end
      elsif @key_state[:cdp] == 2 # rotate
        ang_acc = AMS::ClothWorks.scale_vector(dr, af)
        if kx != 0 && ky == 0 && kz == 0
          @sel_pins_dat[:sid] = 7
        elsif kx == 0 && ky != 0 && kz == 0
          @sel_pins_dat[:sid] = 8
        elsif kx == 0 && ky == 0 && kz != 0
          @sel_pins_dat[:sid] = 9
        else
          @sel_pins_dat[:sid] = nil
        end
      else # translate
        lin_acc = AMS::ClothWorks.scale_vector(dr, af)
        if kx != 0 && ky == 0 && kz == 0
          @sel_pins_dat[:sid] = 1
        elsif kx == 0 && ky != 0 && kz == 0
          @sel_pins_dat[:sid] = 2
        elsif kx == 0 && ky == 0 && kz != 0
          @sel_pins_dat[:sid] = 3
        else
          @sel_pins_dat[:sid] = nil
        end
      end

      lin_damp = AMS::ClothWorks.scale_vector(@sel_pins_dat[:vel], df)
      ang_damp = AMS::ClothWorks.scale_vector(@sel_pins_dat[:omega], df)
      sca_damp = AMS::ClothWorks.scale_vector(@sel_pins_dat[:scale_vel], df)

      @sel_pins_dat[:vel] += lin_acc - lin_damp
      @sel_pins_dat[:omega] += ang_acc - ang_damp
      @sel_pins_dat[:scale_vel] += sca_acc - sca_damp

      @sel_pins_dat[:scale] += AMS::ClothWorks.scale_vector(@sel_pins_dat[:scale_vel], dt)

      pt = AMS::ClothWorks.scale_point(@sel_pins_dat[:vel], dt * AMS::ClothWorks::METER_TO_INCH)
      tt = ::Geom::Transformation.new(pt)

      mag = @sel_pins_dat[:omega].length.to_f
      if mag > AMS::ClothWorks::EPSILON
        nrm = AMS::ClothWorks.scale_vector(@sel_pins_dat[:omega], 1.0 / mag)
        tr = ::Geom::Transformation.rotation(ORIGIN, nrm, mag * dt)
      else
        tr = ::Geom::Transformation.new
      end

      @sel_pins_dat[:tra1] = @sel_pins_dat[:tra1] * tt * tr
      @sel_pins_dat[:tra2] = @sel_pins_dat[:tra1]

    end

    # Transform pins
    s = @sel_pins_dat[:scale]
    @sel_pins.each { |pin, data|
      mat = data[0].to_a
      for i in 0..2
        mat[12 + i] += data[1][i] * (s[i] - 1)
      end
      ntra = ::Geom::Transformation.new(mat)
      pin.transformation = @sel_pins_dat[:tra2] * ntra
    }
  end

  def update_sel_pins_bounds
    return if @sel_pins.empty?

    if @sel_pins_dat[:bb]
      t = @sel_pins_dat[:tra2]
    else
      tf = @first_sel_pin.transformation
      t = Geom::Transformation.new(tf.xaxis, tf.yaxis, tf.zaxis, ORIGIN)
    end
    ti = t.inverse

    sbb = ::Geom::BoundingBox.new
    @sel_pins_dat[:size] = 0.0
    @sel_pins.each { |pin, data|
      e = pin.su_instance
      next unless e.valid?
      lbb = AMS::ClothWorks.get_definition(e).bounds
      rt = ti * e.transformation
      cnr0 = lbb.corner(0).transform(rt)
      cnr7 = lbb.corner(7).transform(rt)
      sbb.add(cnr0)
      sbb.add(cnr7)
      for i in 1..6
        pt = lbb.corner(i).transform(rt)
        sbb.add(pt)
      end
      @sel_pins_dat[:size] += cnr0.distance(cnr7).to_f
    }
    d = sbb.diagonal.to_f
    if d < @sel_pins_dat[:size]
      @sel_pins_dat[:size] = d
    end
    if sbb.valid?
      bb = ::Geom::BoundingBox.new
      bb.add(sbb.min - SEL_PINS_DRAW[:padding])
      bb.add(sbb.max + SEL_PINS_DRAW[:padding])
      corners = Array.new(8)
      for i in 0...8
        corners[i] = bb.corner(i).transform(t)
      end
      @sel_pins_dat[:bb_points] = [
        corners[0], corners[1],
        corners[1], corners[3],
        corners[3], corners[2],
        corners[2], corners[0],

        corners[4], corners[5],
        corners[5], corners[7],
        corners[7], corners[6],
        corners[6], corners[4],

        corners[0], corners[4],
        corners[1], corners[5],
        corners[3], corners[7],
        corners[2], corners[6],
      ]
    else
      @sel_pins_dat[:bb_points] = nil
    end

    unless @sel_pins_dat[:bb]
      # Set gizmo transformation
      #@sel_pins_dat[:tra1] = ::Geom::Transformation.new(sbb.center)
      #@sel_pins_dat[:tra2] = ::Geom::Transformation.new(sbb.center)
      c = sbb.center.transform(t)
      cd = AMS::ClothWorks.get_definition(@first_sel_pin.su_instance)
      if cd.behavior.always_face_camera?
        tz = t.zaxis.normalize
        if (tz.z.to_f < 0.9999995)
          tx = ::Geom::Vector3d.new(-tz.y, tz.x, 0)
        else
          tx = ::Geom::Vector3d.new(tz.z, 0, -tz.x)
        end
        tx.normalize!
        ty = tz.cross(tx).normalize
      else
        tx = t.xaxis
        ty = t.yaxis
        tz = t.zaxis
      end
      @sel_pins_dat[:tra1] = ::Geom::Transformation.new(tx, ty, tz, c)
      @sel_pins_dat[:tra2] = ::Geom::Transformation.new(tx, ty, tz, c)

      @sel_pins_dat[:vel] = ::Geom::Vector3d.new(0,0,0)
      @sel_pins_dat[:omega] = ::Geom::Vector3d.new(0,0,0)
      @sel_pins_dat[:scale] = ::Geom::Vector3d.new(1,1,1)
      @sel_pins_dat[:scale_vel] = ::Geom::Vector3d.new(0,0,0)

      # Store pin transformations
      tinv = @sel_pins_dat[:tra1].inverse
      @sel_pins.reject! { |pin, data|
        next true unless pin.su_instance.valid?
        data[0] = AMS::ClothWorks.extract_matrix_w_factor(tinv * pin.transformation)
        data[1] = pin.su_instance.bounds.center.transform(tinv)
        false
      }
    end

    @sel_pins_dat[:bb] = sbb
  end

  def draw_selected_pins(view)
    return if @sel_pins.empty?

    # Draw bounding box
    view.line_width = SEL_PINS_DRAW[:bounds_width]
    view.line_stipple = SEL_PINS_DRAW[:bounds_stipple]
    view.drawing_color = SEL_PINS_DRAW[:bounds_color]
    view.draw(GL_LINES, @sel_pins_dat[:bb_points])

    c = @sel_pins_dat[:tra2].origin#@sel_pins_dat[:bb].center
    d = @sel_pins_dat[:size]
    cdist = c.distance(view.camera.eye).to_f
    if cdist > AMS::ClothWorks::EPSILON && d / cdist < GIZMO_MIN_DIST_RATIO
      d = cdist * GIZMO_MIN_DIST_RATIO
    end

    bb = AMS::ClothWorks.draw_gizmo(view, @cursor_pos.x, @cursor_pos.y, @sel_pins_dat[:tra2], d, @sel_pins_dat[:sid], GIZMO_OPTS)
    @bb.add(bb)
  end

  def draw(view)
    return unless @world

    if @animation_stopped
      view.animation = self
      @animation_stopped = false
    end

    draw_cloth_meshes(view)
    draw_selected_pins(view)
    draw_pick_and_drag(view)
    ::GC.start

    if @sel_box_pt && !@world.grab_active?
      view.drawing_color = SEL_BOX_COLOR1
      view.line_stipple = @sel_box_pt.x < @cursor_pos.x ? '' : '_'
      view.line_width = 1
      edges = [
        ::Geom::Point3d.new(@sel_box_pt.x, @sel_box_pt.y, 0),
        ::Geom::Point3d.new(@cursor_pos.x, @sel_box_pt.y, 0),
        ::Geom::Point3d.new(@cursor_pos.x, @cursor_pos.y, 0),
        ::Geom::Point3d.new(@sel_box_pt.x, @cursor_pos.y, 0)]
      view.draw2d(GL_LINE_LOOP, edges)
      if AMS::ClothWorks::SU_MAJOR_VERSION > 7
        view.drawing_color = SEL_BOX_COLOR2
        view.draw2d(GL_POLYGON, edges)
      end
    end

    if AMS::ClothWorks.debug?
      view.drawing_color = PIN_BOUNDS_COLOR
      view.line_width = PIN_BOUNDS_LINE_WIDTH
      view.line_stipple = PIN_BOUNDS_LINE_STIPPLE

      @world.pins.each { |pin|
        pin.aligned_box_edges(PIN_BOUNDS_OFFSET).each { |edges|
          view.draw(GL_LINES, edges)
        }
      }
    end

  end

  def swp_on_page_selected(page1, page2, tab1, tab2)
    if @scene_anim_info[:state] == 0
      dtt = @model.options['PageOptions']['TransitionTime']
      tt = (page2.transition_time < 0 ? dtt : page2.transition_time).to_f
      if tt < AMS::ClothWorks::EPSILON
        odt = page2.delay_time
        ott = page2.transition_time
        page2.delay_time = 0
        page2.transition_time = 0
        @model.pages.selected_page = page2
        page2.delay_time = odt
        page2.transition_time = ott
      else
        AMS::Sketchup.activate_scenes_bar_tab(tab2)
        @scene_info[:active] = true
        @scene_info[:data1] = AMS::ClothWorks::SceneData.new()
        @scene_info[:data2] = AMS::ClothWorks::SceneData.new(page2)
        @scene_info[:transition_time] = tt
        @scene_info[:elasted_time] = 0
      end
    end
    1
  end

  def frameChange(from_scene, to_scene, percent_done)
    if @scene_anim_info[:state] == 0
      dtt = @model.options['PageOptions']['TransitionTime']
      tt = (to_scene.transition_time < 0 ? dtt : to_scene.transition_time).to_f
      if tt < AMS::ClothWorks::EPSILON
        if @scene_info[:active]
          @scene_info[:active] = false
          @scene_info[:data1] = nil
          @scene_info[:data2] = nil
          @scene_info[:transition_time] = 0
          @scene_info[:elasted_time] = 0
        end
      else
        @scene_info[:active] = true
        @scene_info[:data1] = AMS::ClothWorks::SceneData.new()
        @scene_info[:data2] = AMS::ClothWorks::SceneData.new(to_scene)
        @scene_info[:transition_time] = tt
        @scene_info[:elasted_time] = 0
      end
    end
  end

  # @return [Boolean]
  def scenes_animating?
    return false if @scene_anim_info[:state] == 0
    et = @scene_anim_info[:ref_time] - AMS::ClothWorks.animate_scenes_delay
    return et >= 0
  end

  # Determine whether the scene transitioning is active.
  # @return [Boolean]
  def scenes_transitioning?
    @scene_info[:active]
  end

  # Stimulate automatic scene transitioning.
  # @note Other settings like, delay and reverse mode, are acquired from the
  #   settings.
  # @param [Integer] state
  #   * 0 - off/stop
  #   * 1 - one way
  #   * 2 - repeat forth and back
  #   * 3 - loop around
  # @return [Boolean] success
  def animate_scenes(state)
    state = AMS::ClothWorks.clamp_int(state.to_i, 0, 3)
    return false if state == @scene_anim_info[:state]
    if state == 0
      @scene_anim_info[:state] = state
      @scene_anim_info[:data].clear
      @scene_anim_info[:transition_time] = 0
      @scene_anim_info[:elasted_time] = 0
      @scene_anim_info[:ref_time] = 0
      @scene_anim_info[:tabs_size] = nil
      @scene_anim_info[:active_tab] = nil
      @scene_anim_info[:tab_dir] = 1
      return true
    elsif @scene_anim_info[:state] != 0
      @scene_anim_info[:state] = state
      @scene_anim_info[:tab_dir] = 1
      return true
    else
      # Obtain default page parameters
      default_delay_time = @model.options['SlideshowOptions']['SlideTime']
      default_transition_time = @model.options['PageOptions']['TransitionTime']
      # Abort any current scene transitioning
      @scene_info[:active] = false
      @scene_info[:data1] = nil
      @scene_info[:data2] = nil
      @scene_info[:transition_time] = 0
      @scene_info[:elasted_time] = 0
      # Gather scene information
      @scene_anim_info[:transition_time] = 0
      tab_index = 0
      tabs_size = @model.pages.size
      @model.pages.each { |page|
        dt = (page.delay_time < 0 ? default_delay_time : page.delay_time).to_f
        tt = (page.transition_time < 0 ? default_transition_time : page.transition_time).to_f
        @scene_anim_info[:data][tab_index] = {
          :sdelay => @scene_anim_info[:transition_time],
          :edelay => @scene_anim_info[:transition_time] + dt,
          :stransition => @scene_anim_info[:transition_time] + dt,
          :etransition => @scene_anim_info[:transition_time] + dt + tt,
          :scene => AMS::ClothWorks::SceneData.new(page),
          :page => page,
          :delay_time => dt,
          :transition_time => tt
        }
        @scene_anim_info[:transition_time] += dt + tt
        tab_index += 1
      }
      # Verify
      if @scene_anim_info[:transition_time] < AMS::ClothWorks::EPSILON
        @scene_anim_info[:transition_time] = 0
        @scene_anim_info[:data].clear
        return false
      end
      # Record additional info
      @scene_anim_info[:state] = state
      @scene_anim_info[:elasted_time] = AMS::ClothWorks.animate_scenes_reversed? ? @scene_anim_info[:transition_time] : 0
      @scene_anim_info[:ref_time] = 0
      @scene_anim_info[:tabs_size] = tabs_size
      @scene_anim_info[:active_tab] = nil
      @scene_anim_info[:tab_dir] = 1
      # Return success
      return true
    end
  end

  # @return [Boolean] success
  def update_scenes_animation
    return false if @scene_anim_info[:state] == 0
    @scene_anim_info[:ref_time] += AMS::ClothWorks::VIEW_UPDATE_TIMESTEP
    return false if @scene_anim_info[:ref_time] < AMS::ClothWorks.animate_scenes_delay
    tab_dir = AMS::ClothWorks.animate_scenes_reversed? ? -@scene_anim_info[:tab_dir] : @scene_anim_info[:tab_dir]
    last_elasted_time = @scene_anim_info[:elasted_time]
    @scene_anim_info[:elasted_time] += AMS::ClothWorks::VIEW_UPDATE_TIMESTEP * tab_dir
    case @scene_anim_info[:state]
      when 1
        if @scene_anim_info[:elasted_time] < 0
          if last_elasted_time == 0
            self.animate_scenes(0)
            return false
          else
            @scene_anim_info[:elasted_time] = 0
          end
        elsif @scene_anim_info[:elasted_time] > @scene_anim_info[:transition_time]
          if last_elasted_time == @scene_anim_info[:transition_time]
            self.animate_scenes(0)
            return false
          else
            @scene_anim_info[:elasted_time] = @scene_anim_info[:transition_time]
          end
        end
      when 2
        if @scene_anim_info[:elasted_time] < 0
          @scene_anim_info[:elasted_time] = @scene_anim_info[:elasted_time].abs
          @scene_anim_info[:tab_dir] = -@scene_anim_info[:tab_dir]
          tab_dir = -tab_dir
        elsif @scene_anim_info[:elasted_time] > @scene_anim_info[:transition_time]
          @scene_anim_info[:elasted_time] = @scene_anim_info[:transition_time] * 2 - @scene_anim_info[:elasted_time]
          @scene_anim_info[:tab_dir] = -@scene_anim_info[:tab_dir]
          tab_dir = -tab_dir
        end
      when 3
        if @scene_anim_info[:elasted_time] < 0
          @scene_anim_info[:elasted_time] += @scene_anim_info[:transition_time]
        elsif @scene_anim_info[:elasted_time] > @scene_anim_info[:transition_time]
          @scene_anim_info[:elasted_time] -= @scene_anim_info[:transition_time]
        end
    else
      self.animate_scenes(0)
      return false
    end
    @scene_anim_info[:data].each { |tab, data|
      unless data[:page].valid?
        self.animate_scenes(0)
        return false
      end
    }
    @scene_anim_info[:data].each { |tab, data|
      if @scene_anim_info[:elasted_time] >= data[:sdelay] && @scene_anim_info[:elasted_time] <= data[:edelay]
        if @scene_anim_info[:active_tab] != tab
          odt = data[:page].delay_time
          ott = data[:page].transition_time
          data[:page].delay_time = 0
          data[:page].transition_time = 0
          @model.pages.selected_page = data[:page]
          #@model.active_view.invalidate
          ::UI.start_timer(0.01, false) {
            m = ::Sketchup.active_model
            if m && m.valid?
              m.active_view.invalidate
            end
          }
          data[:page].delay_time = odt
          data[:page].transition_time = ott
          @scene_anim_info[:active_tab] = tab
        end
        return true
      elsif @scene_anim_info[:elasted_time] >= data[:stransition] && @scene_anim_info[:elasted_time] <= data[:etransition]
        if @scene_anim_info[:active_tab] != tab
          odt = data[:page].delay_time
          ott = data[:page].transition_time
          data[:page].delay_time = 0
          data[:page].transition_time = 0
          @model.pages.selected_page = data[:page]
          ::UI.start_timer(0.01, false) {
            m = ::Sketchup.active_model
            if m && m.valid?
              m.active_view.invalidate
            end
          }
          data[:page].delay_time = odt
          data[:page].transition_time = ott
          @scene_anim_info[:active_tab] = tab
        end
        if @scene_anim_info[:state] == 3
          if tab_dir > 0
            from_tab = tab
            to_tab = tab + 1
            to_tab = 0 if to_tab == @scene_anim_info[:tabs_size]
            from_data = @scene_anim_info[:data][from_tab]
            to_data =  @scene_anim_info[:data][to_tab]
            return true if from_data[:transition_time] < AMS::ClothWorks::EPSILON
            ratio = (@scene_anim_info[:elasted_time] - from_data[:stransition]) / from_data[:transition_time]
          else
            from_tab = tab + 1
            to_tab = tab
            from_tab = 0 if from_tab == @scene_anim_info[:tabs_size]
            from_data = @scene_anim_info[:data][from_tab]
            to_data =  @scene_anim_info[:data][to_tab]
            return true if to_data[:transition_time] < AMS::ClothWorks::EPSILON
            ratio = (to_data[:etransition] - @scene_anim_info[:elasted_time]) / to_data[:transition_time]
          end
        else
          if tab_dir > 0
            from_tab = tab
            to_tab = tab + 1
            return true if to_tab == @scene_anim_info[:tabs_size]
            from_data = @scene_anim_info[:data][from_tab]
            to_data =  @scene_anim_info[:data][to_tab]
            return true if from_data[:transition_time] < AMS::ClothWorks::EPSILON
            ratio = (@scene_anim_info[:elasted_time] - from_data[:stransition]) / from_data[:transition_time]
          else
            from_tab = tab + 1
            to_tab = tab
            return true if from_tab == @scene_anim_info[:tabs_size]
            from_data = @scene_anim_info[:data][from_tab]
            to_data =  @scene_anim_info[:data][to_tab]
            return true if to_data[:transition_time] < AMS::ClothWorks::EPSILON
            ratio = (to_data[:etransition] - @scene_anim_info[:elasted_time]) / to_data[:transition_time]
          end
        end
        ratio = 0.5 * Math.sin(Math::PI * (ratio - 0.5)) + 0.5
        from_data[:scene].transition(to_data[:scene], ratio)
        return true
      end
    }
    return false
  end

  # @return [Boolean] success
  def update_scene_transitioning
    return false unless @scene_info[:active]
    @scene_info[:elasted_time] += AMS::ClothWorks::VIEW_UPDATE_TIMESTEP
    r1 = AMS::ClothWorks.clamp_real(@scene_info[:elasted_time] / @scene_info[:transition_time], 0.0, 1.0)
    r2 = 0.5 * Math.sin(Math::PI * (r1 - 0.5)) + 0.5
    @scene_info[:data1].transition(@scene_info[:data2], r2)
    if r1 >= 1.0
      @scene_info[:active] = false
      @scene_info[:data1] = nil
      @scene_info[:data2] = nil
      @scene_info[:transition_time] = 0
      @scene_info[:elasted_time] = 0
    end
=begin
    cam = @model.active_view.camera
    if @scene_info[:last_pt]
      pt = cam.eye
      if pt.distance(@scene_info[:last_pt]).to_f > 10.0
        @model.entities.add_line([@scene_info[:last_pt], pt])
        @model.entities.add_line(pt, pt + AMS::Geometry.scale_vector(cam.direction, 50.0))
        @scene_info[:last_pt] = pt
      end
    else
      @scene_info[:last_pt] = cam.eye
    end
=end
    return true
  end

end # class AMS::ClothWorks::Simulation