/*
 * ---------------------------------------------------------------------------------------------------------------------
 *
 * Copyright (C) 2019 Anton Synytsia
 *
 * ---------------------------------------------------------------------------------------------------------------------
 */

var g_ms_ie = null;
var g_chrome = null;

String.prototype.contains = function (it) {
  return this.indexOf(it) !== -1;
};

function callback(name, data) {
  if (!data) data = '';
  if (typeof sketchup === 'undefined') {
    window.location.href = 'skp:' + name + '@' + data;
  } else {
    var f = window['sketchup'][name];
    if (f !== undefined) f(data);
  }
}

function init() {
  var cw = $(window).width();
  var ch = $(window).height();
  var bw = 350;
  var bh = 670;
  var data =
    '[' +
    String(cw) +
    ',' +
    String(ch) +
    ',' +
    String(bw) +
    ',' +
    String(bh) +
    ']';

  callback('init', data);
}

function output_to_console(t) {
  callback('output_to_console', String(t));
}

function setup_tabs() {
  var tabs = $('.cw-tabs').children('div');

  tabs.off('click').on('click', function () {
    var content = '#' + this.id + '-content';
    $('.cw-tab-content').css('display', 'none');
    $(content).css('display', 'block');
    tabs.removeClass('cw-tab-active').addClass('cw-tab');
    $(this).removeClass('cw-tab').addClass('cw-tab-active');
  });

  $('.cw-tab-content').css('display', 'none');

  tabs.each(function () {
    if ($(this).hasClass('cw-tab-active')) {
      var content = '#' + this.id + '-content';
      $(content).css('display', 'block');
    }
  });
}

function activate_tab(num) {
  var tabs = $('.cw-tabs').children('div');
  var id = '#tab' + String(num);
  var content_id = id + '-content';

  tabs.removeClass('cw-tab-active').addClass('cw-tab');
  $('.cw-tab-content').css('display', 'none');

  $(id).removeClass('cw-tab').addClass('cw-tab-active');
  $(content_id).css('display', 'block');
}

function eval_string_to_float(str) {
  var num = 0.0;
  try {
    with (Math) {
      num = eval(str);
    }
  } catch (err) {
    return 0.0;
  }
  if (typeof num !== 'number' || num === NaN || num === Infinity) return 0.0;
  return num;
}

function clamp_number(number, min, max) {
  if (number < min) {
    return min;
  } else if (number > max) {
    return max;
  } else {
    return number;
  }
}

function set_slider_value(id, val) {
  var se = document.getElementById('slider' + id);
  var te = document.getElementById('rstext' + id);

  if (se !== null) {
    se.value = String(val);
    val = se.value;
  }
  if (te !== null) {
    te.value = String(val);
  }
}

function enable_slider(id) {
  var se = document.getElementById('slider' + id);
  var te = document.getElementById('rstext' + id);

  se.disabled = false;
  te.disabled = false;
}

function disable_slider(id) {
  var se = document.getElementById('slider' + id);
  var te = document.getElementById('rstext' + id);

  se.disabled = true;
  te.disabled = true;
}

function setup_sliders() {
  var all_sliders = document.getElementsByClassName('cw-slider');
  var i;

  for (i = 0; i < all_sliders.length; ++i) {
    var slider = all_sliders[i];
    if (slider.id.length > 6 && slider.id.substring(0, 6) === 'slider') {
      if (g_ms_ie) {
        slider.onchange = function (evt) {
          evt.preventDefault();
          var id = this.id.substring(6, this.id.length);
          var te = document.getElementById('rstext' + id);
          if (te !== null) {
            te.value = this.value;
          }
          var data = '[:' + this.id + ',' + String(this.value) + ']';
          callback('slider_changing', data);
        };
        slider.onmouseup = function (evt) {
          evt.preventDefault();
          var id = this.id.substring(6, this.id.length);
          var te = document.getElementById('rstext' + id);
          if (te !== null) {
            te.value = this.value;
          }
          var data = '[:' + this.id + ',' + String(this.value) + ']';
          callback('slider_changed', data);
        };
      } else {
        slider.oninput = function (evt) {
          evt.preventDefault();
          var id = this.id.substring(6, this.id.length);
          var te = document.getElementById('rstext' + id);
          if (te !== null) {
            te.value = this.value;
          }
          var data = '[:' + this.id + ',' + String(this.value) + ']';
          callback('slider_changing', data);
        };
        slider.onchange = function (evt) {
          evt.preventDefault();
          var id = this.id.substring(6, this.id.length);
          var te = document.getElementById('rstext' + id);
          if (te !== null) {
            te.value = this.value;
          }
          var data = '[:' + this.id + ',' + String(this.value) + ']';
          callback('slider_changed', data);
        };
      }
    }
  }

  var all_slider_texts = document.getElementsByClassName('cw-slider-text');

  for (i = 0; i < all_slider_texts.length; ++i) {
    var slider_text = all_slider_texts[i];
    if (
      slider_text.id.length > 6 &&
      slider_text.id.substring(0, 6) === 'rstext'
    ) {
      $(slider_text)
        .off('focusout')
        .on('focusout', function () {
          var id = this.id.substring(6, this.id.length);
          var se = document.getElementById('slider' + id);

          if (se !== null) {
            var num = eval_string_to_float(this.value);
            se.value = String(num);
            this.value = se.value;
            var data = '[:slider' + id + ',' + String(se.value) + ']';
            callback('slider_changed', data);
          }
        });

      var id = slider_text.id.substring(6, slider_text.id.length);
      var se = document.getElementById('slider' + id);
      if (se !== null) {
        slider_text.value = se.value;
      }
    }
  }
}

function setup_chosen() {
  $('.cw-chosen-select').chosen('destroy');
  $('.cw-chosen-select').chosen({
    disable_search_threshold: 5,
    no_results_text: 'None matched',
    width: '100%',
  });

  $('.cw-chosen-select')
    .off('change')
    .on('change', function (evt, params) {
      var data = '[:' + this.id + ',"' + params.selected + '"]';
      callback('select_changed', data);
    });
}

function display_help_box(state) {
  $('#help_box').css('display', state ? 'block' : 'none');
  $('#help_box').val('');
}

function setup_checkbox() {
  $('input[type=checkbox], input[type=radio]')
    .off('change')
    .on('change', function () {
      var data = '[:' + this.id + ',' + String(this.checked) + ']';
      callback('check_input_changed', data);
    });
  $('#checkboxE1').on('change', function () {
    display_help_box(this.checked);
  });
}

function setup_buttons() {
  $('#buttonB1')
    .off('click')
    .on('click', function () {
      var data = '{';
      data +=
        ':density=>' + String(document.getElementById('sliderB1').value) + ',';
      data +=
        ':thickness=>' +
        String(parseFloat(document.getElementById('sliderB2').value) * 0.001) +
        ',';
      data +=
        ':friction=>' + String(document.getElementById('sliderB7').value) + ',';
      data +=
        ':shear_stiffness=>' +
        String(document.getElementById('sliderB3').value) +
        ',';
      data +=
        ':bend_stiffness=>' +
        String(document.getElementById('sliderB4').value) +
        ',';
      data +=
        ':damper=>' + String(document.getElementById('sliderB5').value) + ',';
      data +=
        ':scale=>' + String(document.getElementById('sliderB6').value) + ',';
      data +=
        ':drag=>' + String(document.getElementById('checkboxB1').checked) + ',';
      data +=
        ':self_collide=>' +
        String(document.getElementById('checkboxB2').checked);
      data += '}';
      callback('create_material', data);
    });

  $('#buttonB2')
    .off('click')
    .on('click', function () {
      var data = $('#selectB1').val();
      callback('rename_material', data);
    });

  $('#buttonB3')
    .off('click')
    .on('click', function () {
      var data = $('#selectB1').val();
      callback('delete_material', data);
    });

  /*$("input[type=button]").off("click").on("click", function() {
        callback("button_clicked", this.id);
    });*/
}

function setup_help_box() {
  $('a,p,label,input,button').mouseover(function () {
    if (this.title !== '') $('#help_box').val(this.title);
  });
}

function blur_input() {
  if (document.activeElement instanceof HTMLInputElement) {
    document.activeElement.blur();
  }
}

$(document).ready(function () {
  // Browser stuff
  var ua = window.navigator.userAgent;
  var old_ie = ua.indexOf('MSIE ');
  var new_ie = ua.indexOf('Trident/');

  g_ms_ie = old_ie > -1 || new_ie > -1 ? true : false;
  g_chrome = ua.toLowerCase().indexOf('chrome') > -1 ? true : false;

  // Handle errors
  window.onerror = function (msg, url, linenumber) {
    alert('Error: ' + msg + '\nLine: ' + linenumber);
    return true;
  };

  // Handle focus
  $(window).blur(function () {
    blur_input();
  });

  // Setup tabs
  setup_tabs();

  // Setup sliders
  setup_sliders();

  // Setup chosen
  setup_chosen();

  // Setup check and radio input
  setup_checkbox();

  // Setup buttons
  setup_buttons();

  // Setup help box
  setup_help_box();

  // Initialize
  window.setTimeout(init, 0);
});
